import os
import math
import pickle
from numpy import random
import numpy as np
from datetime import datetime

from ase.utils import basestring
from ase.io.trajectory import Trajectory
from asap3.nanoparticle_mc.cluster import Cluster
from asap3.nanoparticle_mc.clusteratom import ClusterAtom
from asap3.nanoparticle_mc import data

class AtomMonteCarlo:
    def __init__(self, cluster=None, debug=0, asap=None):
        self.debug = debug
        if asap is not None:
            self.use_asap = True
            self.asap_calc = asap
        else:
            self.use_asap = False

        if not isinstance(cluster, Cluster):
            raise Warning('The cluster is not a valid Cluster instance.')

        if cluster.symmetry != 'fcc':
            raise Warning('Can only work with fcc structures.')

        #Set the structual parameters
        self.cluster = cluster
        self.size = len(self.cluster)

        self.neighbor_positions = (data.lattice[cluster.symmetry]['neighbor_positions'] * 
                                   self.cluster.lattice_constant)
        self.neighbor_mapping = data.lattice[cluster.symmetry]['neighbor_mapping']
        self.neighbor_count = data.lattice[cluster.symmetry]['neighbor_count']
        self.type_count = data.lattice[cluster.symmetry]['type_count']
        self.type_names = data.lattice[cluster.symmetry]['type_names']
        self.type_numbers = data.lattice[cluster.symmetry]['type_numbers']
        self.type_data = data.lattice[cluster.symmetry]['type_data']

        self.arrays = {}

        #Generate a list with vacant positions and their nearest neighbors
        self.cluster.make_neighborlist()

        positions = self.get_positions()
        neighbors = self.get_neighbors()

        for i in self.surface_indexes():
            self.add_vacancies(positions[i], neighbors[i], i)

        if self.debug > 2:
            if self.check_vacancies():
                raise Warning('Something is wrong, read the message above')

    def run(self, steps=1000, temp=1000, txt=None, amc=None, trajectory=None):
        #Open the output textfile
        if txt is not None:
            if not isinstance(txt, basestring):
                raise Warning('You must specify a valid filename.')
            f = open(txt, 'w')
        else:
            f = None

        #Initialize the dataobject
        if amc is not None:
            amcdata = AtomMonteCarloData(amc,
                                         self.cluster.atomic_number,
                                         self.cluster.symmetry,
                                         self.cluster.lattice_constant,
                                         self.size,
                                         temp)
            amcdata.set_dimensions(self.type_count + 1, self.neighbor_count + 1)
        else:
            amcdata = None

        #Open the trajectoryfile
        if trajectory is not None:
            traj = Trajectory(trajectory, 'w', self.cluster)
            #traj.write_energy = False
            traj.write_forces = False
            traj.write_stress = False
            traj.write_momenta = False
            traj.write()

        #Print header to outputfile
        if isinstance(f, file):
            print(('Monte Carlo Sumulation on Cluster Surface Atoms \n' +
                         'Date: %s\n' % datetime.now().strftime("%A, %d %b %Y %H:%M:%S") +
                         'Material: %s\n' % data.chemical_symbols[self.cluster.atomic_number] +
                         'Lattice constant: %s\n' % self.cluster.lattice_constant +
                         'Cluster size: %i\n' % self.size +
                         'Multiplicity: %i\n' % self.cluster.multiplicity +
                         'Temperature: %i K\n' % temp +
                         'Thermal energy: %.5f eV\n' % (data.kB * temp) +
                         'Iteration  Atom          Position                ' +
                         'Desination           Energy    Accept  Probability'), file=f)

        #Set initial values for the simulation
        old_energy = self.cluster.get_potential_energy()
        acceptances = 0

        if amcdata is not None:
            amcdata.append((old_energy,
                            0.0, #Releaxed energy not found yet
                            self.cluster.get_diameter(),
                            self.cluster.multiplicity,
                            1,
                            np.bincount(self.get_types()),
                            np.bincount(self.get_coordinations()),
                            np.bincount(self.get_vacant_types()),
                            np.bincount(self.get_vacant_coordinations()),
                           ))

        #Print initial energy to outputfile
        if isinstance(f, file):
            print('%6i %70.4f' % (0, old_energy), file=f)

        # Check for ASAP calculator, and enable optimization
        if self.use_asap:
            auxcluster = self.cluster.copy()
            auxcluster.set_calculator(self.asap_calc)
            auxcluster.get_potential_energy()
            
        #Monte Carlo simulering
        for i in range(1, steps + 1):
            #Chose a move
            surface_i, vacant_i = self.chose_move_old()

            if self.use_asap:
                new_atom = auxcluster[surface_i]
                new_atom.set_position(self.get_vacant_positions()[vacant_i])
                new_energy = auxcluster.get_potential_energy()
                # Move atom back.  It is moved forward again if accepted.
                #new_atom.set_position(self.cluster[surface_i].position)
            else:
                #Make new cluster with change and calculate energy
                new_cluster = self.cluster.copy()
                new_cluster.calc = self.cluster.calc
                new_atom = ClusterAtom(new_cluster.atomic_number)
                new_atom.position = self.get_vacant_positions()[vacant_i]
                new_atom.neighbors = self.get_vacant_neighbors()[vacant_i]
                new_atom.type = self.get_vacant_types()[vacant_i]
                new_atom.coordination = self.get_vacant_coordinations()[vacant_i]
                new_cluster.move_atom(surface_i, new_atom)
                new_energy = new_cluster.get_potential_energy()

            #Calculate the energy change and accept or reject the change
            energy_change = new_energy - old_energy
            accept = False

            if energy_change < 0:
                p = 1.0
                accept = True
            else:
                p = math.exp(- energy_change / (data.kB * temp))
                pc = random.random()
                if pc < p: accept = True

            #Print info on MC move til outputfile
            if isinstance(f, file):
                old_position = self.cluster[surface_i].position
                new_position = new_atom.position

                print(('%6i %8i   (%6.2f%7.2f%7.2f)   (%6.2f%7.2f%7.2f) %11.4f' %  
                             (i, surface_i, 
                              old_position[0], old_position[1], old_position[2], 
                              new_position[0], new_position[1], new_position[2], 
                              new_energy)), end=' ', file=f)

            if accept:
                old_energy = new_energy
                acceptances += 1
                self.move_atom(surface_i, vacant_i)

                if isinstance(f, file):
                    print('%7s %10.2f' % ('Yes', p), file=f)

                if amcdata is not None:
                    amcdata.append((old_energy,
                                    0.0, #Releaxed energy not found yet
                                    self.cluster.get_diameter(),
                                    self.cluster.multiplicity,
                                    1,
                                    np.bincount(self.get_types()),
                                    np.bincount(self.get_coordinations()),
                                    np.bincount(self.get_vacant_types()),
                                    np.bincount(self.get_vacant_coordinations()),
                                   ))

                if trajectory is not None:
                    traj.write(self.cluster)
            else:
                if self.use_asap:
                    # Move atom back
                    new_atom.set_position(self.cluster[surface_i].position)
                if isinstance(f, file):
                    print('%7s %10.2f' % ('No', p), file=f)

                if amcdata is not None:
                    amcdata.add_count(-1, 1)

        #Print footer to the outputfile
        if isinstance(f, file):
            print('\nDate: %s' % datetime.now().strftime("%A, %d %b %Y %H:%M:%S"), file=f)
            print('Acceptances ratio: %.2f' % (acceptances * 1.0 / steps), file=f)

        #Save the data object
        if amcdata is not None:
            amcdata.write()

        #Close the trajectory
        if trajectory is not None:
            traj.close()

    def chose_move(self):
        #Choose a surface atom to move and where to move it
        #coordinations = np.arange(1, 12)
        coordinations = np.array([1,1,2,2,3,3,4,4,5,5,6,7,8,9,10,11])
        indexes = np.empty(0)
        while len(indexes) == 0:
            coordination = self.choice(coordinations)
            mask = (self.get_coordinations() == coordination)
            indexes = np.arange(len(mask))[mask]
        surface_i = self.choice(indexes)

        #coordinations = np.arange(1, 12)
        coordinations = np.array([1,2,3,4,5,6,7,8,9,10,11])
        indexes = np.empty(0)
        self_neighbor = True
        while True:
            while len(indexes) == 0:
                coordination = self.choice(coordinations)
                mask = (self.get_vacant_coordinations() == coordination)
                indexes = np.arange(len(mask))[mask]
            vacant_i = self.choice(indexes)
            #The atom must not move to a site where its only neighbor is it self
            #If just one other neighbor is found then its fine => break
            neighbors = self.get_vacant_neighbors()[vacant_i]
            if ((neighbors != -1) * (neighbors != surface_i)).any():
                break

        return surface_i, vacant_i

    def chose_move_old(self):
        #Choose a surface atom to move and where to move it
        surface_i = self.choice(self.surface_indexes())

        self_neighbor = True
        while self_neighbor:
            vacant_i = self.choice(self.vacant_indexes())
            #The atom must not move to a site where its only neighbor is it self
            #If just one other neighbor is found then its fine => break
            for n in self.get_vacant_neighbors()[vacant_i]:
                if n != -1 and n != surface_i:
                    self_neighbor = False
                    break

        return surface_i, vacant_i

    def resize(self, size=None, method="random"):
        if method == "random":
            if size > self.size: #Add
                adding_atoms = size - self.size 
                added_atoms = 0
                coordination = 12

                while True:
                    coordination -= 1
                    coordinations = self.get_vacant_coordinations()
                    indexes = np.arange(len(coordinations))[coordinations >= coordination]
                    difference = adding_atoms - added_atoms

                    if self.debug:
                        print ("Found %i vacancies with coordination %i." %
                               (len(indexes), coordination))

                    if len(indexes) > difference:
                        for i in range(difference):
                            n = self.choice(range(len(indexes)))
                            self.add_atom(indexes[n])
                            added_atoms += 1
                            indexes = indexes - (indexes > indexes[n]).astype(int)
                            indexes = np.delete(indexes, n)
                    else:
                        for n in range(len(indexes)):
                            self.add_atom(indexes[n])
                            added_atoms += 1
                            indexes = indexes - (indexes > indexes[n]).astype(int)

                    if added_atoms == adding_atoms:
                        break

            elif size < self.size: #Remove
                removing_atoms = self.size - size
                removed_atoms = 0
                coordination = 0

                while True:
                    coordination += 1
                    coordinations = self.get_coordinations()
                    indexes = np.arange(len(coordinations))[coordinations <= coordination]
                    difference = removing_atoms - removed_atoms

                    if self.debug:
                        print ("Found %i atoms with coordination %i." %
                               (len(indexes), coordination))

                    if len(indexes) > difference:
                        for i in range(difference):
                            n = self.choice(range(len(indexes)))
                            self.remove_atom(indexes[n])
                            removed_atoms += 1
                            indexes = indexes - (indexes > indexes[n]).astype(int)
                            indexes = np.delete(indexes, n)
                    else:
                        for n in range(len(indexes)):
                            self.remove_atom(indexes[n])
                            removed_atoms += 1
                            indexes = indexes - (indexes > indexes[n]).astype(int)

                    if removed_atoms == removing_atoms:
                        break

        elif method == "cluster":
            raise Warning('The method is not supported yet.')

        self.size = len(self.cluster)

        if self.debug > 2:
            check_vacancies = self.check_vacancies()
            check_neighbors = self.check_neighbors()

            if check_vacancies or check_neighbors:
                raise Warning('Something is wrong, read message above')

    #Manipulation functions for atoms
    def add_atom(self, index):
        """ Adds an atom to the vacant position given by "index". The
        vacant position is removed and new vacancies are added around
        the atom. The atoms neighbors neighborlists are updated with
        the atoms new index.
        """

        #Set up the new atom with position, neighbors, type and coordination
        atom = ClusterAtom(self.cluster.atomic_number)
        atom.position = self.get_vacant_positions()[index]
        atom.neighbors = self.get_vacant_neighbors()[index]
        atom.coordination = self.get_vacant_coordinations()[index]
        atom.type = self.get_vacant_types()[index]

        #Add the atom
        self.cluster.add_atom(atom)

        #Print that an atom has been added
        if self.debug:
            position = atom.position
            print (">> Adding atom %i at vacancy %i (%2.2f %2.2f %2.2f):" %
                   (atom.index, index, position[0], position[1], position[2]))

        #Remove old vacacy and add new vacancies
        self.remove_vacancy(index)
        self.add_vacancies(atom.position, atom.neighbors, atom.index)

    def remove_atom(self, index):
        """ Removes the atom given by "index". Its position is added as
        a vacancy and vacancies only associated with the atom is removed.
        Neighbor references are updated, because the atoms gets new 
        indexes.
        """

        #Remove and save the atom (neighborlist indexes are shifted due to the delete)
        atom = self.cluster.remove_atom(index)

        #Print that an atom is removed
        if self.debug:
            position = atom.position
            print (">> Removing atom %i (%2.2f %2.2f %2.2f):" % 
                   (index, position[0], position[1], position[2]))
            
        #Remove old vacancies
        self.remove_vacancies(index)

        #Subtract one from all neighbor references that is greater than "index"
        vacant_neighbors = self.get_vacant_neighbors()
        vacant_neighbors = vacant_neighbors - (vacant_neighbors > index).astype(int)
        self.set_vacant_neighbors(vacant_neighbors)

        #Add new vacancy
        self.add_vacancy(atom.position, atom.neighbors, atom.coordination, atom.type)

    def move_atom(self, i, vacant_i):
        """ Moves an atom given by "i" to the vacant positon given by
        "vacant_i". The old position is added as vacancy and the new 
        position is removed. Atoms and vacancies around the old and 
        new position are updated.
        """

        #Save the old position, neighbors, type and coordination
        old_atom = self.cluster[i]
        old_atom.cut_reference_to_atoms()

        #Set up the new atom with position, neighbors, type and coordination
        new_atom = ClusterAtom(self.cluster.atomic_number)
        new_atom.position = self.get_vacant_positions()[vacant_i]
        new_atom.neighbors = self.get_vacant_neighbors()[vacant_i]
        new_atom.coordination = self.get_vacant_coordinations()[vacant_i]
        new_atom.type = self.get_vacant_types()[vacant_i]

        self.cluster.move_atom(i, new_atom)

        #Print that an atom is moved
        if self.debug:
            old_position = old_atom.position
            new_position = self.cluster[i].position
            print (">> Moving atom %i (%2.2f %2.2f %2.2f) " %
                   (i, old_position[0], old_position[1], old_position[2]) +
                   "to vacancy %i (%2.2f %2.2f %2.2f):" % 
                   (vacant_i, new_position[0], new_position[1], new_position[2]))
            
        #Add the old and remove the new position
        self.remove_vacancy(vacant_i)
        self.add_vacancy(old_atom.position,
                         old_atom.neighbors,
                         old_atom.coordination,
                         old_atom.type)

        #Update vacancies around the moved atom old and new position
        self.remove_vacancies(i)
        self.add_vacancies(self.cluster[i].position, self.cluster[i].neighbors, i)

        if self.debug > 2:
            check_vacancies = self.check_vacancies()
            check_neighbors = self.check_neighbors()

            if check_vacancies or check_neighbors:
                raise Warning('Something is wrong, read message above')

    #Manipulation functions for vacancies
    def add_vacancy(self, position, neighbors, coordination, type):
        """Adds a single vacancy, without updating any lists."""
        vacant_positions = self.get_vacant_positions()
        vacant_neighbors = self.get_vacant_neighbors()
        vacant_coordinations = self.get_vacant_coordinations()
        vacant_types = self.get_vacant_types()

        if self.debug:
            print ("Adding vacacy %i (%2.2f %2.2f %2.2f)" %
                   (len(vacant_positions), position[0], position[1], position[2]))

        vacant_positions = np.append(vacant_positions,
                                     position.reshape(1, 3),
                                     axis=0)
        vacant_neighbors = np.append(vacant_neighbors,
                                     neighbors.reshape(1, self.neighbor_count),
                                     axis=0)
        vacant_coordinations = np.append(vacant_coordinations, 
                                         coordination)
        vacant_types = np.append(vacant_types, 
                                 type)

        self.set_vacant_positions(vacant_positions)
        self.set_vacant_neighbors(vacant_neighbors)
        self.set_vacant_coordinations(vacant_coordinations)
        self.set_vacant_types(vacant_types)

    def remove_vacancy(self, index):
        """Removes a single vacancy, without updating any lists."""
        vacant_positions = self.get_vacant_positions()
        vacant_neighbors = self.get_vacant_neighbors()
        vacant_coordinations = self.get_vacant_coordinations()
        vacant_types = self.get_vacant_types()

        if self.debug:
            position = vacant_positions[index]
            print ("Removing vacacy %i (%2.2f %2.2f %2.2f)" %
                   (index, position[0], position[1], position[2]))

        vacant_positions = np.delete(vacant_positions, index, axis=0)
        vacant_neighbors = np.delete(vacant_neighbors, index, axis=0)
        vacant_coordinations = np.delete(vacant_coordinations, index)
        vacant_types = np.delete(vacant_types, index)

        self.set_vacant_positions(vacant_positions)
        self.set_vacant_neighbors(vacant_neighbors)
        self.set_vacant_coordinations(vacant_coordinations)
        self.set_vacant_types(vacant_types)

    def add_vacancies(self, position, neighbors, index):
        """Adds and updates vacancies associated with the atom 
        at "position" indexed with "index". It will only add a
        vacancy if it is not pressent in the list of vacancies
        already."""

        vacant_positions = self.get_vacant_positions()
        vacant_neighbors = self.get_vacant_neighbors()
        vacant_coordinations = self.get_vacant_coordinations()
        vacant_types = self.get_vacant_types()

        for i, n in enumerate(neighbors):
            if n < 0:
                vacant_position = position + self.neighbor_positions[i] 
                vacant_append = True

                for j, vp in enumerate(vacant_positions):
                    if (np.abs(vp - vacant_position) < 1e-3).all():
                        vacant_neighbors[j][self.neighbor_mapping[i]] = index
                        vacant_coordinations[j] += 1
                        vacant_types[j] = self.get_atom_type(vacant_neighbors[j])
                        vacant_append = False

                if vacant_append:
                    if self.debug:
                        print ("Adding vacacy %i (%2.2f %2.2f %2.2f)" %
                               (len(vacant_positions),
                                vacant_position[0], 
                                vacant_position[1], 
                                vacant_position[2]))

                    vacant_positions = np.append(vacant_positions,
                                                 vacant_position.reshape(1,3),
                                                 axis=0)
                    vacant_neighbors = np.append(vacant_neighbors,
                                                 -np.ones((1,self.neighbor_count), dtype=int),
                                                 axis=0)
                    vacant_neighbors[-1][self.neighbor_mapping[i]] = index
                    vacant_coordinations = np.append(vacant_coordinations, 1)
                    vacant_types = np.append(vacant_types, 
                                             self.get_atom_type(vacant_neighbors[-1]))

        self.set_vacant_positions(vacant_positions)
        self.set_vacant_neighbors(vacant_neighbors)
        self.set_vacant_coordinations(vacant_coordinations)
        self.set_vacant_types(vacant_types)

    def remove_vacancies(self, index):
        """Removes and updates vacancies associated with the 
        atom indexed "index". It will only remove vacancies
        if they are not associated with other atoms."""

        vacant_positions = self.get_vacant_positions()
        vacant_neighbors = self.get_vacant_neighbors()
        vacant_coordinations = self.get_vacant_coordinations()
        vacant_types = self.get_vacant_types()

        i = 0
        end = len(vacant_neighbors)
        while i < end:
            vacant_remove = True

            for j, n in enumerate(vacant_neighbors[i]):
                if n >= 0 and n != index:
                    vacant_remove = False
                if n == index:
                    vacant_neighbors[i][j] = -1
                    vacant_coordinations[i] -= 1
                    vacant_types[i] = self.get_atom_type(vacant_neighbors[i])

            if vacant_remove:
                if self.debug:
                    position = vacant_positions[i]
                    print ("Removing vacacy %i (%2.2f %2.2f %2.2f)" %
                           (i, position[0], position[1], position[2]))

                end -= 1
                vacant_positions = np.delete(vacant_positions, i, axis=0)
                vacant_neighbors = np.delete(vacant_neighbors, i, axis=0)
                vacant_coordinations = np.delete(vacant_coordinations, i)
                vacant_types = np.delete(vacant_types, i)
            else:
                i += 1

        self.set_vacant_positions(vacant_positions)
        self.set_vacant_neighbors(vacant_neighbors)
        self.set_vacant_coordinations(vacant_coordinations)
        self.set_vacant_types(vacant_types)

    #Functions to check for faults in the neighbor lists
    def check_neighbors(self):
        positions = self.get_positions()
        neighbors = self.get_neighbors()
        coordinations = self.get_coordinations()
        types = self.get_types()
        trouble = False

        for i, n in enumerate(neighbors):
            for j, m in enumerate(n):
                found = False
                p1 = positions[i] + self.neighbor_positions[j]

                for k, p2 in enumerate(positions):
                    if (np.abs(p1 - p2) < 1e-3).all():
                        found = True
                        if m < 0:
                            print('Atom %i is missing neighbor atom %i at %s' % (i, k, p2))
                            trouble = True
                        elif m != k:
                            print ('Atom %i is having atom %i as ' +
                                   'neighbor instead of atom %i' % (i, m, k))
                            trouble = True

                if m >= 0 and found is False:
                    trouble = True
                    print('Atom %i does not have atom %i as neighbor' % (i, m))

            if self.get_atom_coordination(n) != coordinations[i]:
                trouble = True
                print('Atom %i coordination does not match' % i)

            if self.get_atom_type(n) != types[i]:
                trouble = True
                print('Atom %i type does not match' % i)

        return trouble

    def check_vacancies(self):
        vacant_positions = self.get_vacant_positions()
        vacant_neighbors = self.get_vacant_neighbors()
        positions = self.get_positions()
        trouble = False
        
        for i, p1 in enumerate(vacant_positions):
            for j, p2 in enumerate(positions):
                if (np.abs(p1 - p2) < 1e-3).all():
                    trouble = True
                    print("Vacancy %i is identical to %i at %s" % (i, j, p1))

        for i, n in enumerate(vacant_neighbors):
            for j, m in enumerate(n):
                p1 = vacant_positions[i] + self.neighbor_positions[j]

                for k, p2 in enumerate(positions):
                    if (np.abs(p1 - p2) < 1e-3).all():
                        if m < 0:
                            print('Vacancy %i is missing neighbor atom %i at %s' % (i, k, p2))
                            trouble = True
                        elif m != k:
                            print ('Vacancy %i is having atom %i as ' % (i, m) +
                                   'neighbor instead of atom %i' % k)
                            trouble = True

        return trouble

    #Some useful functions
    def choice(self, a):
        return a[random.randint(len(a))]

    def surface_indexes(self):
        """Returns the indexes for all the surface atoms (coordination < 12)."""
        mask = self.get_coordinations() < 12
        return np.arange(len(self.cluster))[mask]

    def vacant_indexes(self):
        """Returns the indexes for the vacancies."""
        return np.arange(len(self.get_vacant_positions()))

    def get_atom_coordination(self, neighbors):
        neighbors = np.array(neighbors)
        neighbors = neighbors[neighbors >= 0]
        return len(neighbors)

    def get_atom_type(self, neighbors):
        neighbors = np.array(neighbors)
        typename = tuple((neighbors >= 0).astype(int))
        if typename in self.type_names:
            return self.type_data[self.type_numbers[typename]]['type']
        else:
            return 0

    #Positions (edited on self.cluster)
    def get_positions(self, index=None):
        return self.cluster.get_positions()

    def set_positions(self, positions):
        return self.cluster.set_positions(positions)

    #Neighbors
    def get_neighbors(self):
        return self.cluster.get_neighbors()

    def set_neighbors(self, neighbors):
        self.cluster.set_neighbors(neighbors)

    #Coordination
    def get_coordinations(self):
        return self.cluster.get_coordinations()

    def set_coordinations(self, coordinations):
        self.cluster.set_coordinations(coordinations)

    #Types
    def get_types(self):
        return self.cluster.get_types()

    def set_types(self, types):
        self.cluster.set_types(types)

    #Vacant positions
    def get_vacant_positions(self):
        if 'vacant_positions' not in self.arrays:
            return np.empty((0, 3), dtype=float)
        else:
            return self.arrays['vacant_positions'].copy()

    def set_vacant_positions(self, positions):
        self.arrays['vacant_positions'] = np.array(positions, dtype=float)

    #Vacant neighbors
    def get_vacant_neighbors(self):
        if 'vacant_neighbors' not in self.arrays:
            return np.empty((0, self.neighbor_count), dtype=int)
        else:
            return self.arrays['vacant_neighbors'].copy()

    def set_vacant_neighbors(self, neighbors):
        self.arrays['vacant_neighbors'] = np.array(neighbors, dtype=int)

    #Vacant coordinations
    def get_vacant_coordinations(self):
        if 'vacant_coordinations' not in self.arrays:
            return np.empty(0, dtype=int)
        else:
            return self.arrays['vacant_coordinations'].copy()

    def set_vacant_coordinations(self, coordinations):
        self.arrays['vacant_coordinations'] = np.array(coordinations, dtype=int)

    #Vacant types
    def get_vacant_types(self):
        if 'vacant_types' not in self.arrays:
            return np.empty(0, dtype=int)
        else:
            return self.arrays['vacant_types'].copy()

    def set_vacant_types(self, types):
        self.arrays['vacant_types'] = np.array(types, dtype=int)

class AtomMonteCarloData:
    def __init__(self, filename=None, symbol=None, symmetry=None, latticeconstant=None,
                 size=None, temp=None):
        if isinstance(symbol, basestring):
            self.atomic_number = data.atomic_numbers[symbol]
        else:
            self.atomic_number = symbol

        self.filename = filename
        self.symmetry = symmetry
        self.lattice_constant = latticeconstant
        self.size = size
        self.temp = temp

        self.arrays = {}
        self.arrays['energy'] = np.empty(0, dtype=float)
        self.arrays['energyrelax'] = np.empty(0, dtype=float)
        self.arrays['size'] = np.empty(0, dtype=float)
        self.arrays['multi'] = np.empty(0, dtype=int)
        self.arrays['counts'] = np.empty(0, dtype=int)
        self.arrays['types'] = np.empty(0, dtype=int)
        self.arrays['coordinations'] = np.empty(0, dtype=int)
        self.arrays['vtypes'] = np.empty(0, dtype=int)
        self.arrays['vcoordinations'] = np.empty(0, dtype=int)

    def set_dimensions(self, type_count, coord_count):
        self.arrays['types'] = self.arrays['types'].reshape((0, type_count))
        self.arrays['coordinations'] = self.arrays['coordinations'].reshape((0, coord_count))
        self.arrays['vtypes'] = self.arrays['vtypes'].reshape((0, type_count))
        self.arrays['vcoordinations'] = self.arrays['vcoordinations'].reshape((0, coord_count))

    def __len__(self):
        return len(self.arrays['energy'])

    def __getitem__(self, index=-1):
        energy = self.arrays['energy'][index]
        energyrelax = self.arrays['energyrelax'][index]
        size = self.arrays['size'][index]
        counts = self.arrays['counts'][index]
        multi = self.arrays['multi'][index]
        types = self.arrays['types'][index]
        coordinations = self.arrays['coordinations'][index]
        vtypes = self.arrays['vtypes'][index]
        vcoordinations = self.arrays['vcoordinations'][index]
        
        return (energy, energyrelax, size, multi, counts,
                types, coordinations, vtypes, vcoordinations)

    def __delitem__(self, index=-1):
        mask = np.ones(len(self), bool)
        mask[index] = False

        for name, a in self.arrays.items():
            self.arrays[name] = a[mask]

    def __setitem__(self, index, values):
        try:
            if values[0] is not None:
                self.arrays['energy'][index] = values[0]

            if values[1] is not None:
                self.arrays['energyrelax'][index] = values[1]

            if values[2] is not None:
                self.arrays['size'][index] = values[2]

            if values[3] is not None:
                self.arrays['multi'][index] = values[3]

            if values[4] is not None:
                self.arrays['counts'][index] = values[4]

            if values[5] is not None:
                a = self.arrays['types']
                b = np.zeros(a.shape[1], a.dtype)
                b[:len(values[5])] = np.array(values[5], a.dtype)
                self.arrays['types'][index] = b

            if values[6] is not None:
                a = self.arrays['coordinations']
                b = np.zeros(a.shape[1], a.dtype)
                b[:len(values[6])] = np.array(values[6], a.dtype)
                self.arrays['coordinations'][index] = b

            if values[7] is not None:
                a = self.arrays['vtypes']
                b = np.zeros(a.shape[1], a.dtype)
                b[:len(values[7])] = np.array(values[7], a.dtype)
                self.arrays['vtypes'][index] = b

            if values[8] is not None:
                a = self.arrays['vcoordinations']
                b = np.zeros(a.shape[1], a.dtype)
                b[:len(values[8])] = np.array(values[8], a.dtype)
                self.arrays['vcoordinations'][index] = b
        except:
            print(values)
            raise

    def append(self, values):
        for name, a in self.arrays.items():
            b = np.zeros((a.shape[0] + 1,) + a.shape[1:], a.dtype)
            b[:a.shape[0]] = a
            self.arrays[name] = b

        self[-1] = values

    def extend(self, other):
        if not isinstance(other, AtomMonteCarloData):
            raise Warning('The added object is not an AtomMonteCarloData instance!')

        if len(self) == 0:
            self.set_dimensions(other.get_types().shape[1],
                                other.get_coordinations().shape[1])

        for name, a in self.arrays.items():
            b = other.arrays[name]
            if a.shape[1:] != b.shape[1:]:
                raise Warning('Array {} has different shapes: {} != {}'.format(name,
                                                                           a.shape[1:],
                                                                           b.shape[1:]))
            c = np.zeros((a.shape[0] + b.shape[0],) + a.shape[1:], a.dtype)
            c[:a.shape[0]] = a
            c[a.shape[0]:] = b
            self.arrays[name] = c

    def add_count(self, index, value):
        self.arrays['counts'][index] += value

    def filter_mask(self, mask):
        """Function that filters the data with the given mask."""

        if len(mask) != len(self):
            raise Warning(f'The mask has a wrong size: {len(mask)} != {len(self)}')

        for name, a in self.arrays.items():
            self.arrays[name] = a[mask]

    def filter_energy(self, excess):
        """Function that filters the data leaving only entries with
        energy lower than 'excess'."""

        e = self.get_energies()
        de = (e - e.min())
        mask = de < excess
        self.filter_mask(mask)

    def filter_relaxed(self):
        """Function that filters the data leaving only relaxed entries."""

        mask = (np.abs(self.get_relaxed_energies()) > 1e-3)
        self.filter_mask(mask)

    def get_energies(self):
        return self.arrays['energy'].copy()

    def get_relaxed_energies(self):
        return self.arrays['energyrelax'].copy()

    def get_sizes(self):
        return self.arrays['size'].copy()

    def get_multiplicities(self):
        return self.arrays['multi'].copy()

    def get_counts(self):
        return self.arrays['counts'].copy()

    def get_counts_mask(self):
        mask = np.empty(0, int)
        for i, c in enumerate(self.get_counts()):
            mask_add = np.ones(c, int) * i
            mask_old = mask.copy()
            mask = np.zeros(len(mask_add) + len(mask_old), int)
            mask[:len(mask_old)] = mask_old
            mask[len(mask_old):] = mask_add
        return mask

    def get_types(self):
        return self.arrays['types'].copy()

    def get_coordinations(self):
        return self.arrays['coordinations'].copy()

    def get_vacant_types(self):
        return self.arrays['vtypes'].copy()

    def get_vacant_coordinations(self):
        return self.arrays['vcoordinations'].copy()

    #Functions to write and read the data on the disk
    def write(self, filename=None, zipped=True, backup=True):
        if filename is None:
            filename = self.filename

        if not isinstance(filename, basestring):
            raise Warning('You must specify a valid filename.')

        if backup:
            if os.path.isfile(filename):
                os.rename(filename, filename + '.bak')
            elif os.path.isfile(filename + ".gz"):
                os.rename(filename + ".gz", filename + '.gz.bak')

        d = {'atomicnumber': self.atomic_number,
             'symmetry': self.symmetry,
             'latticeconstant': self.lattice_constant,
             'size': self.size,
             'temp': self.temp}

        if zipped:
            f = os.popen('gzip > ' + filename + '.gz', 'wb')
        else:
            f = open(filename, 'wb')

        f.write('AtomMonteCarlo')
        pickle.dump(d, f)
        pickle.dump(self.arrays, f)
        f.close()

    def read(self, filename=None):
        if filename is None:
            filename = self.filename

        if not isinstance(filename, basestring):
            raise Warning('You must specify a valid filename.')

        if os.path.isfile(filename):
            if filename[-3:] == '.gz':
                filename = filename[:-3]
                f = os.popen("zcat " + filename + ".gz", "rb")
            else:
                f = open(filename, 'rb')
        elif os.path.isfile(filename + '.gz'):
            f = os.popen("zcat " + filename + '.gz', "rb")
        else:
            raise Warning(f'The file {filename} do not exist.')

        self.filename = filename

        try:
            if f.read(len('AtomMonteCarlo')) != 'AtomMonteCarlo':
                raise Warning('This is not a compatible file.')
            d = pickle.load(f)
            self.arrays = pickle.load(f)
        except EOFError:
            raise Warinig('Bad file.')

        f.close()

        if 'multiplicity' in d:
            self.multiplicity = d['multiplicity']
            self.arrays['multi'] = np.ones(len(self), float) * d['multiplicity']

        if 'energyrelax' not in self.arrays:
            self.arrays['energyrelax'] = np.zeros(len(self))

        if 'size' not in self.arrays:
            self.arrays['size'] = np.zeros(len(self))

        self.atomc_number = d['atomicnumber']
        self.lattice_constant = d['latticeconstant']
        self.size = d['size']
        self.temp = d['temp']




from asap3 import EMT
from ase.lattice.cubic import FaceCenteredCubic
import numpy as np
from asap3.test.pytest_markers import ReportTest, serial
import pytest



@serial
def test_changeperiodic():

    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(6,6,6),
                            symbol="Cu", pbc=(1,1,0))
    atoms.calc = EMT()
    ecorrect = atoms.get_potential_energy()

    atoms2 = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(6,6,6),
                            symbol="Cu", pbc=(0,0,0))
    atoms2.calc = EMT()
    atoms2.get_potential_energy()
    atoms2.set_pbc((1,1,0))
    e1= atoms2.get_potential_energy()
    ReportTest("PBC (0,0,0) -> (1,1,0) correct", e1, ecorrect, 0.001)

    atoms3 = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(6,6,6),
                            symbol="Cu", pbc=(1,1,1))
    atoms3.calc = EMT()
    atoms3.get_potential_energy()
    atoms3.set_pbc((1,1,0))
    e3= atoms3.get_potential_energy()
    ReportTest("PBC (0,0,0) -> (1,1,0) correct", e3, ecorrect, 0.001)




    atoms.set_pbc((0,1,0))
    atoms.get_potential_energy()
    atoms.set_pbc((1,1,0))
    e2 = atoms.get_potential_energy()
    ReportTest("e2 correct", e2, ecorrect, 0.001)

    print("Setting pbc")
    atoms.set_pbc((0,1,0))
    print("Calculating energy")
    dummy =  atoms.get_potential_energy()
    assert np.fabs(dummy - ecorrect) > 1.0
    atoms.set_pbc((1,1,0))
    e3 = atoms.get_potential_energy()
    ReportTest("e3 correct", e3, ecorrect, 0.001)

    atoms.set_pbc((1,1,0))
    atoms.set_positions(atoms.get_positions())
    e4 = atoms.get_potential_energy()
    ReportTest("e4 correct", e4, ecorrect, 0.001)

    print("Repeating tests with an atom outside the unit cell.")
    for coordinate in (0,1,2):
        print("Using coordinate number", coordinate)

        atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], size=(6,6,6),
                                symbol="Cu", pbc=(1,1,0))
        r = atoms.get_positions()
        uc = atoms.get_cell()
        r[-1,coordinate] = uc[coordinate,coordinate] * 1.51
        atoms.set_positions(r)
        atoms.calc = EMT()
        ecorrect = atoms.get_potential_energy()

        atoms.set_pbc((0,1,0))
        atoms.set_pbc((1,1,0))
        e2 = atoms.get_potential_energy()
        ReportTest("e2 correct", e2, ecorrect, 0.001)

        atoms.set_pbc((0,1,0))
        dummy =  atoms.get_potential_energy()
        assert np.fabs(dummy - ecorrect) > 1.0
        atoms.set_pbc((1,1,0))
        e3 = atoms.get_potential_energy()
        ReportTest("e3 correct", e3, ecorrect, 0.001)

        atoms.set_pbc((1,1,0))
        atoms.set_positions(atoms.get_positions())
        e4 = atoms.get_potential_energy()
        ReportTest("e4 correct", e4, ecorrect, 0.001)



"Test the RawRadialDistributionFunction function."

from numpy import *
from asap3 import *
from asap3.analysis.rdf import RadialDistributionFunction
from ase.lattice.cubic import FaceCenteredCubic
from ase.lattice.compounds import NaCl
from asap3.mpi import world
from asap3.test.pytest_markers import ReportTest, serial

import pytest


ismaster = world.rank == 0
isparallel = world.size != 1

# testtypes: latticeconst, maxRDF, bins, useEMT
testtypes = ((3.6, 6.001, 100, False),
             (3.7, 6.001, 100, True),
             (3.6, 15.001, 100, False),
             (3.65, 15.001, 100, True))

@pytest.mark.parametrize('latconst,maxrdf,nbins,withemt', testtypes)
def test_rdf_fcc(latconst, maxrdf, nbins, withemt, cpulayout):
    if ismaster:
        atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], symbol="Cu",
                              size=(10,10,10), latticeconstant=latconst)
        if isparallel:
            atoms = atoms.repeat(cpulayout)
    else:
        atoms = None
    if isparallel:
        atoms = MakeParallelAtoms(atoms, cpulayout)
    natoms = atoms.get_global_number_of_atoms()

    if withemt:
        atoms.calc = EMT()
        print(atoms.get_potential_energy())

    rdf = RadialDistributionFunction(atoms, maxrdf, nbins, verbose=True)
    z = atoms.get_atomic_numbers()[0]

    globalrdf = rdf.get_rdf()
    localrdf = rdf.get_rdf(elements=(z,z))

    print(globalrdf)

    ReportTest("Local and global RDF are identical",
               max(abs(globalrdf - localrdf)), 0.0, 1e-6)
        
    shellpop = [12, 6, 24, 12, 24, -1]
    shell = [sqrt(i+1.0)/sqrt(2.0) for i in range(6)]
    print(shell)
    print(shellpop)
    n = 0

    dr = maxrdf/nbins

    for i in range(nbins):
        if (i+1)*dr >= shell[n] * latconst:
            if shellpop[n] == -1:
                print("Reached the end of the test data")
                break
            rho = natoms / atoms.get_volume()
            expected = shellpop[n] / (4 * pi * ((i+0.5) * dr)**2 * dr * rho)
            maxerr = shellpop[n] / (8 * pi * ((i+0.5) * dr)**3 * rho) * 0.02
            print("Shell", n+1, globalrdf[i], expected, maxerr, (globalrdf[i] - expected)/maxerr)
            ReportTest(("Shell %d (%d)" % (n+1, i)), globalrdf[i],
                       expected, maxerr)
            n += 1
        else:
            ReportTest(("Between shells (%d)" % (i,)), globalrdf[i], 0, 0)

# testtypes: latticeconst, maxRDF, bins, useEMT, save
testtypes2 = ((5.1, 6.001, 100, False, True),
             (5.2, 6.001, 100, True, False),
             (5.1, 15.001, 100, False, False),
             (5.15, 15.001, 100, True, False)
             )

@serial
@pytest.mark.xfail  # From old disabled test RDF2.py
@pytest.mark.parametrize('latconst,maxrdf,nbins,withemt,save', testtypes2)
def test_rdf_NaCl(latconst, maxrdf, nbins, withemt, save, in_tmp_dir):
    atoms = NaCl(directions=[[1,0,0],[0,1,0],[0,0,1]],
                 symbol=("Cu", "Au"), size=(10,10,10),
                 latticeconstant=latconst, debug=0)
    natoms = len(atoms)
    ReportTest("Number of atoms", natoms, 8000, 0)

    if withemt:
        atoms.calc = EMT()
        print(atoms.get_potential_energy())

    rdf = RadialDistributionFunction(atoms, maxrdf, nbins)
    if save:
        rdf.output_file("RDF2-rdf")
    z = atoms.get_atomic_numbers()[0]

    globalrdf = rdf.get_rdf()
    localrdf = zeros(globalrdf.shape, float)
    for i in (29,79):
        for j in (29,79):
            localrdf += rdf.get_rdf(elements=(i,j))

    ReportTest("Local and global RDF are identical",
               min( globalrdf == localrdf), 1, 0)
        
    shellpop = [6, 12, 8, 6, 24, -1]
    shell = [sqrt(i+1.0)/2.0 for i in range(6)]
    n = 0

    dr = maxrdf/nbins

    for i in range(nbins):
        if (i+1)*dr >= shell[n] * latconst:
            if shellpop[n] == -1:
                print("Reached the end of the test data")
                break
            rho = len(atoms) / atoms.get_volume()
            expected = shellpop[n] / (4 * pi * ((i+0.5) * dr)**2 * dr * rho)
            maxerr = shellpop[n] / (8 * pi * ((i+0.5) * dr)**3 * rho)
            print("expected, error", expected, maxerr, globalrdf[i]-expected)
            ReportTest(("Shell %d (%d)" % (n+1, i)), globalrdf[i],
                       expected, maxerr)
            n += 1
        else:
            ReportTest(("Between shells (%d)" % (i,)), globalrdf[i], 0, 0)

    if save:
        rdf = RadialDistributionFunction.load("RDF2-rdf0000.rdf")

        newglobalrdf = rdf.get_rdf()
        ReportTest("Saved and original RDF have same length",
                   len(newglobalrdf), len(globalrdf), 0)
        for i in range(len(globalrdf)):
            ReportTest("Saved and global RDF at position %d" % (i,),
                       newglobalrdf[i], globalrdf[i], 0)
            
        localrdf = zeros(globalrdf.shape, float)
        for i in (29,79):
            for j in (29,79):
                x = rdf.get_rdf(elements=(i,j))
                localrdf += x

        ReportTest("Saved local and global RDF are identical",
                   min( globalrdf == localrdf), 1, 0)

        os.unlink("RDF2-rdf0000.rdf")
    




import numpy as np

def CohesiveEnergy(atoms):
    e_full = atoms.get_potential_energy() / len(atoms)

    atoms2 = atoms[[0,]]
    atoms2.set_calculator(atoms.calc)
    e_reduced = atoms2.get_potential_energy()

    return e_reduced - e_full

def VacancyEnergy(atoms):
    e_full = atoms.get_potential_energy()

    atoms2 = atoms.copy()
    atoms2.set_calculator(atoms.calc)
    atoms2.pop()
    e_reduced = atoms.get_potential_energy()

    return e_reduced - e_full * len(atoms) / (len(atoms) + 1)

def HeatOfFormation(atoms):
    return None

def HeatOfSolution(atoms):
    return None



from asap3 import *
from asap3.md.verlet import VelocityVerlet
from ase.lattice.cubic import FaceCenteredCubic
from asap3.io.trajectory import Trajectory
from ase.parallel import world
from asap3.md.velocitydistribution import MaxwellBoltzmannDistribution, Stationary
from asap3.constraints import FixAtoms
from ase.io import read
import os

import pytest


@pytest.mark.core
def test_Fixatoms_trajectory(cpulayout, in_tmp_dir):

    ismaster = world.rank == 0
    isparallel = world.size != 1

    delete = True
    filename = "fixatoms.traj"

    if ismaster:
        initial = FaceCenteredCubic(size=(10,10,10), symbol="Cu", pbc=(1,0,0))
    else:
        initial = None
    if isparallel:
        atoms = MakeParallelAtoms(initial, cpulayout)
        atoms.set_constraint(FixAtoms(mask=(atoms.get_ids() < 5)))
    else:
        atoms = initial.copy()
        atoms.set_constraint(FixAtoms(indices=range(5)))

    atoms.calc = EMT()
    MaxwellBoltzmannDistribution(atoms, temperature_K=5000)
    Stationary(atoms)

    dyn = VelocityVerlet(atoms, 3*units.fs)
    traj = Trajectory(filename, "w", atoms)
    dyn.attach(traj, interval=10)
    dyn.run(50)

    e = atoms.get_potential_energy()

    atoms2 = read(filename)
    e2 = atoms2.get_potential_energy()

    world.barrier()
    if delete and world.rank == 0:
        os.remove(filename)
        
    world.barrier()



import sys
import numpy as np
import numbers

from ase.utils import basestring
from asap3.Tools.ParameterOptimization.Quantities import QuantityCalc

def ParameterPerformance(calculator, quantities, latticeconstants, txt=None,
                         debug=False, noprint=False):
    """
    Calculates the performance of a calculator based on some quantities.

    Parameters
    -----------
    calculator: Calculator object.

    quantities: List of quantities which should be calculated. The list should
    be on the form [(quantity_name, structure, elements, target), ...].

    txt: Output file. If none standard output is used.
    """
    quantitycalc = QuantityCalc(latticeconstants, calculator, debug=debug)

    if txt is None:
        f = sys.stdout
    else:
        f = open(txt, 'a')

    results = []
    for quan in quantities:
        name = quan[0]
        structure = quan[1]
        elements = quan[2]
        target = quan[3]
        valuefunc = getattr(quantitycalc, 'get_' + name)
        if isinstance(elements, (basestring, numbers.Integral)):
            elements = (elements,)
        else:
            elements = tuple(elements)
        value = valuefunc(structure, elements)

        results.append((name, structure, elements, target, value))

    if noprint:
        return results
    
    f.write("Potential Performance\n" +
            "%-45s  %-7s  " % ("Quantity, Structure:", "Target") +
            "%8s    %-9s\n" % ("Value", "Deviation") +
            "-" * 77 + "\n")
    devsum = 0
    for (name, structure, elements, target, value) in results:
        if len(elements) > 1:
            struct = "%s%s-%s" % (elements + (structure,))
        else:
            struct = "%s-%s" % (elements + (structure,))
        name = f"{name}, {struct}:"
        dev = value / target * 100 - 100
        devsum += np.abs(dev)
        f.write("%-45s  %7.3f  " % (name, target) +
                f"{value:10.6f}  {dev:8.2f}%\n")
    f.write(f"{devsum / len(results):76.2f}%\n\n")
    #f.close()





"""Asap utility functions

This module defines the following functions:
PrintVersion
"""

__docformat__ = "restructuredtext en"


from asap3.Internal.Builtins import _asap, get_version, get_short_version
from asap3 import __file__ as _asapfile
import numpy as np
import ase
import sys
import os

def print_version(level = 0):
    """Print the version number of the loaded version of Asap.

    If the optional argument is 1, also prints the pathnames of the
    most important files.
    """
    try:
        compiledfile = _asap.__file__
    except AttributeError:
        compiledfile = "<built-in>"
    print(get_version())
    if level >= 1:
        pv = sys.version.split('\n')
        print("  Python module: ", _asapfile)
        print("  C++ module:    ", compiledfile)
        print("  ase module:    ", ase.__file__)
        print()
        print("  Python version:", pv[0])
        for pvline in pv[1:]:
            print("                 ", pvline)
        print("  numpy version: ", np.version.version)
        print("  ase version:   ", ase.__version__)

def DebugOutput(filename, stdout=1, nomaster=False, sync=True):
    """Debugging output on each node goes to a different file.

    Redirect stderr to a different file on each node.  The filename should
    contain %d, which is replaced by the node number.  The file is opened
    with minimal buffering, and stderr (standard error) is redirected to
    it (also for C/C++ extensions).  If the optional argument stdout is
    true (the default), Python's sys.stdout is also redirected to the
    same file.  Standard output for C/C++ extensions is never touched.

    This is mainly useful for parallel simulations.
    """
    if stdout:
        sys.stdout = sys.stderr
    try:
        import asap3.mpi
        node = asap3.mpi.world.rank
    except (AttributeError, ImportError):
        node = 0
    if nomaster and node == 0:
        return
    flag = os.O_WRONLY|os.O_CREAT|os.O_TRUNC
    if sync:
        flag = flag|os.O_SYNC
    newerror = os.open((filename % (node,)), flag, 0o660)
    try:
        os.dup2(newerror, sys.stderr.fileno())
    except:
        sys.stderr.write("Cannot redirect sys.stderr as it is fake.")
    # This Python file must NOT go away.  Attach it to the sys module.
    sys._AsapStandardError = newerror

def print_memory(txt, a=None):
    import asap3.mpi
    procfile = open("/proc/self/status")
    vmsize = vmpeak = vmdata = vmrss = -1
    for line in procfile:
        words = line.split()
        if words[0] == "VmSize:":
            vmsize = int(words[1])
        elif words[0] == "VmPeak:":
            vmpeak = int(words[1])
        elif words[0] == "VmData:":
            vmdata = int(words[1])
        elif words[0] == "VmRSS:":
            vmrss = int(words[1])
    print("Memory [proc %d '%s']: %d MB total (%d MB peak, %d MB data, %d MB rss)" % (
        asap3.mpi.world.rank, txt, (vmsize+512) / 1024,
        (vmpeak+512) / 1024, (vmdata+512) / 1024, (vmrss+512)/1024)) #, file=sys.stderr)        
    procfile.close()
    if a is not None:
        memory_usage(a)

def memory_usage(obj, total=True):
    """Print the memory usage of some kinds of objects.

    Supported objects are: atoms, EMT calculators and neighbor lists.
    """
    mem = 0
    if hasattr(obj, "arrays"):
        mem += _memory_usage_atoms(obj)
        try:
            calc = obj.calc
        except AttributeError:
            calc = None
        if calc is not None:
            mem += memory_usage(calc, total=False)
    elif hasattr(obj, "print_memory"):
        mem += obj.print_memory()
    else:
        print("*MEM* Memory usage of this object is not supported:", obj)
        return 0
    if total:
        print("*MEM* Total %d MB." % (mem,))
    return mem

def _memory_usage_atoms(atoms):
    arr = atoms.arrays
    mem = 0
    nat = len(atoms)
    nvar = 0
    megabyte = 1024*1024
    for k in arr.keys():
        mem += arr[k].size * arr[k].itemsize
        nvar += 1
    gmem = 0
    gvar = 0
    if hasattr(atoms, "ghosts"):
        arr = atoms.ghosts
        for k in arr.keys():
            gmem += arr[k].size * arr[k].itemsize
            gvar += 1
    mem = (mem + gmem + megabyte/2) / megabyte
    gmem = (gmem + megabyte/2) / megabyte
    print("*MEM* Atoms %d MB.  [ %d atoms, %d arrays, %d gh_arr of %d MB ]" % (
        mem, nat, nvar, gvar, gmem))
    return mem



from asap3 import *
from ase.lattice.cubic import FaceCenteredCubic
from asap3.testtools import ReportTest
from asap3.mpi import world
from asap3.Internal.ParallelListOfAtoms import ParallelAtoms
import numpy as np

#DebugOutput("migration%d.log", nomaster=True)

def pot():
    #return EMT2013(PtY_parameters)
    return EMT()

#set_verbose(1)
master = world.rank == 0
if world.size == 1:
    cpulayout = None
elif world.size == 2:
    cpulayout = [2,1,1]
elif world.size == 3:
    cpulayout = [1,3,1]
elif world.size == 4:
    cpulayout = [2,1,2]
    
if master:
    atoms0 = FaceCenteredCubic(symbol='Pt', size=(15,15,30))
else:
    atoms0 = None
    
atoms0 = MakeParallelAtoms(atoms0, cpulayout)
atoms0.calc = pot()

print("*********** FIRST FORCE CALCULATION ************", file=sys.stderr)
print("len(atoms) =", len(atoms0), "   no. atoms =", atoms0.get_global_number_of_atoms(), file=sys.stderr)
f0 = atoms0.get_forces()
perturbation = 0.01 * np.random.standard_normal(atoms0.get_positions().shape)
r = atoms0.get_positions() + perturbation
atoms0.set_positions(r)
print("*********** SECOND FORCE CALCULATION ************", file=sys.stderr)

f1 = atoms0.get_forces()

print("*********** COPYING ATOMS **************", file=sys.stderr)
#atoms0.set_calculator(None)  # Cannot copy parallel calculators
atoms2 = ParallelAtoms(cpulayout, atoms0.comm, atoms0, distribute=False)
atoms2.calc = pot()
print("*********** THIRD FORCE CALCULATION ************", file=sys.stderr)
f2 = atoms2.get_forces()

#maxdev = abs(f2 - f1).max()
#print maxdev
#ReportTest("Max error 1:", maxdev, 0.0, 1e-6)

#ReportTest.Summary()

print("No crashes - success !!", file=sys.stderr)



from numpy import *
#from Scientific.Statistics import *
from asap3 import *
from asap3.md.npt import NPT
from asap3.md.langevin import Langevin
from asap3.io.trajectory import Trajectory
from ase.lattice.cubic import FaceCenteredCubic
from asap3.testtools import ReportTest
from asap3.mpi import world
from ase import units
import sys
import numpy as np

T_goal = 300 # K
p_goal = 1 # GPa
bulk = 137 # Gpa
cp = 24.43 # J / (K * mol)
step1 = 25

#set_verbose(1)

np.random.seed([4314523, world.size, world.rank])

if len(sys.argv) == 1 or sys.argv[0].endswith("TestAll.py"):
    out1 = "testNPT-1.out"
    out2 = "testNPT-2.out"
    ttime = 25
    ptime = 75
elif len(sys.argv) == 3:
    ttime = float(sys.argv[1])
    ptime = float(sys.argv[2])
    out1 = "testNPT-%d-%d-1.out" % (ttime, ptime)
    out2 = "testNPT-%d-%d-2.out" % (ttime, ptime)
else:
    raise RuntimeError("Expected zero or two parameter to the script!" + str(sys.argv))

ismaster = world.rank == 0
if world.size == 2:
    cpulayout = [2,1,1]
elif world.size == 4:
    cpulayout = [2,2,1]
elif world.size == 8:
    cpulayout = [2,2,2]
else:
    raise RuntimeError("Cannot run on %i CPUs" % (world.size,))

if ismaster:
    print_version(1)

if not hasattr(NPT, '_npt_version'):
    print("Skipping test: NP dynamics does not work in parallel with this old version of ASE.")
else:
    if ismaster:
        atoms = FaceCenteredCubic(size=(30,15,15), symbol="Cu", pbc=True)
    else:
        atoms = None
    atoms = MakeParallelAtoms(atoms, cpulayout)
    atoms.calc = EMT()
    print("Number of atoms:", atoms.get_global_number_of_atoms())

    print("Heating to %d K using Langevin" % T_goal)
    lgv = Langevin(atoms, 5 * units.fs, temperature_K=2*T_goal,
                       friction=0.05)

    while atoms.get_kinetic_energy() < 1.5 * atoms.get_global_number_of_atoms() * T_goal * units.kB:
        lgv.run(5)
        T = atoms.get_kinetic_energy() / (1.5 * atoms.get_global_number_of_atoms() * units.kB)
        print(f"Temperature is now {T:.2f} K")
    print("Desired temperature reached!")

    lgv.set_temperature(temperature_K=T_goal)

    for i in range(2):
        lgv.run(20)
        s = atoms.get_stress()
        p = -(s[0] + s[1] + s[2])/3.0 / units.GPa
        T = atoms.get_kinetic_energy() / (1.5 * atoms.get_global_number_of_atoms() * units.kB)
        print(f"Pressure is {p:f} GPa, desired pressure is {p_goal:f} GPa (T = {T:.2f} K)")
        dv = (p - p_goal) / bulk
        print("Adjusting volume by", dv)
        cell = atoms.get_cell()
        atoms.set_cell(cell * (1.0 + dv/3.0))

    T = atoms.get_kinetic_energy() / (1.5 * atoms.get_global_number_of_atoms() * units.kB)
    print(f"Temperature is now {T:.2f} K")

    stressstate = array([-2, -1, 0, 0, 0, 0])*p_goal*units.GPa
    dyn = NPT(atoms, 5 * units.fs, temperature_K=T_goal,
                  externalstress=stressstate,
                  ttime=ttime*units.fs,
                  pfactor=(ptime*units.fs)**2 * bulk * units.GPa)
    traj = Trajectory("NPT-atoms.traj", "w", atoms)
    #dyntraj = ParallelHooverNPTTrajectory("NPT-dyn-traj.nc", dyn, interval = 50)
    dyn.attach(traj, interval=50)
    #dyn.Attach(dyntraj)

    out = open(out1, "w")

    temp = []
    pres = []
    vol = []
    for i in range(step1):
        dyn.run(5)
        T = atoms.get_kinetic_energy() / (1.5 * atoms.get_global_number_of_atoms() * units.kB)
        s = atoms.get_stress() / units.GPa
        p = -(s[0] + s[1] + s[2])/3.0
        out.write("%5.2f %5.2f %7.5f      %7.5f %7.5f %7.5f %7.5f %7.5f %7.5f\n" %
                  ((dyn.get_time(), T, p)+tuple(s)))
        out.flush()
        cell = atoms.get_cell()
        v = cell[0,0] * cell[1,1] * cell[2,2]
        temp.append(T)
        pres.append(p)
        vol.append(v)
        if i % 5 == 0:
            print(i,"/",step1)

    del atoms, dyn, traj
    #del dyntraj




from pathlib import Path
from contextlib import contextmanager
import os
from asap3.test.pytest_markers import _openkimmodel

import pytest

try:
    from asap3.mpi import world
except ImportError:
    from ase.parallel import world
from ase.parallel import broadcast


# Add markers here instead of in pytest.ini as that file is not always found by pytest
def pytest_configure(config):
    config.addinivalue_line(
        "markers", "core: very important test of core functionality"
    )
    config.addinivalue_line(
        "markers", "slow: test is particularly slow"
    )


### Skip tests marked as slow unless --runslow is given on command line
### See https://docs.pytest.org/en/latest/example/simple.html#control-skipping-of-tests-according-to-command-line-option
def pytest_addoption(parser):
    parser.addoption(
        "--slow", action="store_true", default=False, help="run slow tests"
    )

def pytest_collection_modifyitems(config, items):
    if config.getoption("--slow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --slow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


# Fixtures

@pytest.fixture(scope='session')
def datadir():
    test_basedir = Path(__file__).parent
    return test_basedir / 'testdata'

@pytest.fixture(scope='session')
def cpulayout():
    if world.size == 1:
        return None
    elif world.size == 2:
        return [2,1,1]
    elif world.size == 3:
        return [1,3,1]
    elif world.size == 4:
        return [2,1,2]
    elif world.size == 6:
        return [1,2,3]
    elif world.size == 8:
        return [2,2,2]
    else:
        raise ValueError('Test suit should run on 1, 2, 3, 4, 6 or 8 cores.')

@pytest.fixture(scope='session')
def multicpulayout():
    if world.size == 1:
        return [None]
    elif world.size == 2:
        return [(2,1,1), (1,2,1), (1,1,2)]
    elif world.size == 3:
        return [(3,1,1), (1,1,3)]
    elif world.size == 4:
        return [(1,2,2), (2,1,2), (2,2,1)]
    elif world.size == 8:
        return [(2,2,2), (1,2,4)]
    else:
        raise ValueError("Cannot run on %d CPUs." % (worldsize,))

@pytest.fixture(scope='session')
def openkimmodel():
    'Default OpenKIM model for tests'
    return _openkimmodel

# Run some modules in a temporary folder - copied from GPAW

@contextmanager
def execute_in_tmp_path(request, tmp_path_factory):
    if world.rank == 0:
        # Obtain basename as
        # * request.function.__name__  for function fixture
        # * request.module.__name__    for module fixture
        basename = getattr(request, request.scope).__name__
        path = tmp_path_factory.mktemp(basename)
    else:
        path = None
    path = broadcast(path)
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(cwd)

@pytest.fixture(scope='function')
def in_tmp_dir(request, tmp_path_factory):
    """Run test function in a temporary directory."""
    with execute_in_tmp_path(request, tmp_path_factory) as path:
        yield path


@pytest.fixture(scope='module')
def module_tmp_path(request, tmp_path_factory):
    """Run test module in a temporary directory."""
    with execute_in_tmp_path(request, tmp_path_factory) as path:
        yield path



#!/usr/bin/env python

"""OfficialTiming.py - Asap timing script.

Reports timing of various types of dynamics in usec/atom/timestep.
"""

from numpy import *
from asap3 import *
from ase.lattice.cubic import FaceCenteredCubic
from asap3.testtools import ReportTest
from asap3.md.velocitydistribution import MaxwellBoltzmannDistribution
from asap3 import mpi
from asap3.md.npt import NPT
from ase import units
from io import StringIO
import sys
from subprocess import getoutput
import time

possibleLayouts = [(1,1,2), (1,2,2), (2,2,2), (2,2,3), (2,2,4), (2,3,3), 
                   (2, 3, 4), (2, 4, 4), (2, 4, 5), (3,3,3), (3,3,4), (3,4,4), (4,4,4), 
                   (4,4,5), (4,5,5), (4,6,6), (6,6,6)]
systemSizes = [((6,6,7), 1500), ((13,14,14), 150), ((29,29,30), 15), 
               ((63,63,63), 8)]
##systemSizes = [((6,6,7), 100), ((13,14,14), 10)]

if len(sys.argv) > 1:
    postfix = "-" + sys.argv[1]
else:
    postfix = ""
logfilename = f"paralleltiming{postfix}.log"
bigtablename = f"paralleltiming{postfix}-big.table"
fasttablename = f"paralleltiming{postfix}-fast.table"

ismaster = mpi.world.rank == 0

class Logger(StringIO):
    def __init__(self, filename, silent=False):
        StringIO.__init__(self)
        self.logfile = filename
        self.silent = silent
    def write(self, x):
        if not self.silent:
            sys.stdout.write(x)
        StringIO.write(self, x)
    def close(self):
        if ismaster:
            outfile = open(self.logfile, "a")
            outfile.write(self.getvalue())
            outfile.close()
        StringIO.close(self)

logger = Logger(logfilename, silent=(not ismaster))
bigtable = Logger(bigtablename, silent=True)
fasttable = Logger(fasttablename, silent=True)

def SerialTiming(name, func, timesteps, natoms):
    if ismaster:
        print("Preparing timing")
    func(2)
    if ismaster:
        print("Running serial timing:", name)
    mpi.world.barrier()
    startcpu, startwall = time.clock(), time.time()
    func(timesteps)
    cpu, wall = time.clock() - startcpu, time.time() - startwall
    fraction = 100.0 * cpu/wall
    if ismaster:
        print("Time:", cpu, "CPU seconds  ", wall, "Wall seconds") 
    wall = mpi.world.sum(wall)
    wall *= 1e6 / (timesteps * natoms * mpi.world.size)
    fraction = mpi.world.min(fraction)
    if ismaster:
        logger.write("Serial %s: %.2f usec/atom/timestep Wall-time  (%.1f%%)\n"
                     % (name, wall, fraction))
    return wall

def ParallelTiming(name, func, timesteps, natoms, serialtime, summary):
    if ismaster:
        print("Preparing timing")
    func(2)
    if ismaster:
        print("Running parallel timing:", name)
    mpi.world.barrier()
    startcpu, startwall = time.clock(), time.time()
    func(timesteps)
    mpi.world.barrier()
    cpu, wall = time.clock() - startcpu, time.time() - startwall
    if ismaster:
        print("Time:", cpu, "CPU seconds  ", wall, "Wall seconds")
    wall *= 1e6 / (timesteps * natoms) * mpi.world.size
    efficiency = 100.0 * serialtime / wall
    if ismaster:
        logger.write("Parallel %s: %.2f usec/atom/timestep Wall-time. Efficiency: %.1f%%\n"
                     % (name, wall, efficiency))
        summary.write("%5.2f %3d%% " % (wall, efficiency))

def temperature(atoms):
    return 2.0/3.0 * atoms.get_kinetic_energy() / len(atoms) / units.kB

def MakeCu(size, T=300):
    if ismaster:
        print("Preparing", T, "K Copper system.")
    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                              symbol='Cu', size=size, debug=0)
    atoms.set_calculator(EMT())
    MaxwellBoltzmannDistribution(atoms, 2*T * units.kB)
    #dyn = VelocityVerlet(atoms, 5*units.fs)
    dyn = Langevin(atoms, 5*units.fs, T*units.kB, 0.05)
    dyn.run(50)
    if ismaster:
        print("Done.  Temperature =", temperature(atoms), \
              "K.  Number of atoms: ", len(atoms))
    return atoms


host = getoutput("hostname")
when = time.strftime("%a %d %b %Y %H:%M", time.localtime(time.time()))
asapversion = get_version()

random.seed([42+mpi.world.rank, 12345*mpi.world.rank])

layouts = {}
for l in possibleLayouts:
    layouts[ l[0]*l[1]*l[2] ] = l

if ismaster:
    print("Running on %d processors with %s as master" % (mpi.world.size, host))

cpuLayout = layouts[mpi.world.size]


# Make sure that CPU speed is revved up.
dummy = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]], symbol='Cu',
                          size=(10,10,10), debug=0)
dummy.set_calculator(EMT())
dyn = Langevin(dummy, 5*units.fs, 300*units.kB, 0.05)
dyn.run(10)
del dummy

logger.write("\n\nRunning on %d processors with %s as master (%s)\n"
             % (mpi.world.size, host, when))
modelname = "Unknown CPU model name"
cpumhz = "Unknown CPU speed"
try:
    lines = open("/proc/cpuinfo").readlines()
    for line in lines:
        if line[:10] == "model name":
            modelname = line
            break
    for line in lines:
        if line[:7] == "cpu MHz":
            cpumhz = line
            break
except:
    print("Cannot get CPU info from /proc/cpuinfo")

logger.write(cpumhz)
logger.write(modelname)
logger.write(asapversion+"\n")

if ismaster:
    bigtable.write("%-5d " % mpi.world.size)
    fasttable.write("%-5d " % mpi.world.size)

for size, steps in systemSizes:
    atoms = MakeCu(size)
    atomspercpu = len(atoms)
    positions = atoms.get_positions()
    basis = atoms.get_cell()
    atomicnumber = int(atoms.get_atomic_numbers()[0])

    # Run the serial timings
    MaxwellBoltzmannDistribution(atoms, 300 * units.kB)
    name = "Verlet%d" % (atomspercpu,)
    atoms.set_calculator(EMT())
    dyn = VelocityVerlet(atoms, 5*units.fs)
    verlet = SerialTiming(name, dyn.run, steps, atomspercpu)
    name = "Langevin%d" % (atomspercpu,)
    dyn = Langevin(atoms, 5*units.fs, 300*units.kB, 0.001)
    langevin = SerialTiming(name, dyn.run, steps, atomspercpu)
    #name = "NPT%d" % (atomspercpu,)
    #dyn = NPT(atoms, 5*units.fs, 300*units.kB, 0, 25*units.fs,
    #          (75*units.fs)**2 * 140*units.GPa)
    #npt = SerialTiming(name, dyn.run, steps, atomspercpu)

    # A distributed atoms.Repeat
    n = mpi.world.rank
    gridpos = (n % cpuLayout[0], (n / cpuLayout[0]) % cpuLayout[1],
               n / (cpuLayout[0] * cpuLayout[1]))
    newpos = positions + dot(array(gridpos), basis)[newaxis,:]
    atoms = Atoms(positions=newpos,
                  cell = dot(array(cpuLayout), basis),
                  pbc = (1,1,1))
    atoms.set_atomic_numbers(atomicnumber*ones(len(atoms),int))
    atoms = MakeParallelAtoms(atoms, cpuLayout)

    # Run the parallel timings (big systems)
    MaxwellBoltzmannDistribution(atoms, 300 * units.kB)
    name = "Verlet%d-big" % (atomspercpu,)
    atoms.set_calculator(EMT())
    dyn = VelocityVerlet(atoms, 5*units.fs)
    ParallelTiming(name, dyn.run, 4*steps, atoms.get_global_number_of_atoms(),
                   verlet, bigtable)
    name = "Langevin%d-big" % (atomspercpu,)
    dyn = Langevin(atoms, 5*units.fs, 300*units.kB, 0.001)
    ParallelTiming(name, dyn.run, 4*steps, atoms.get_global_number_of_atoms(),
                   langevin, bigtable)
    #name = "NPT%d-big" % (atomspercpu,)
    #dyn = NPT(atoms, 5*units.fs, 300*units.kB, 0, 25*units.fs,
    #          (75*units.fs)**2 * 140*units.GPa)
    #ParallelTiming(name, dyn.run, 4*steps, atoms.get_global_number_of_atoms(),
    #               npt, bigtable)

    # Make parallel systems without repeating
    if ismaster:
        atoms = Atoms(positions=positions,
                      cell = basis,
                      pbc = (1,1,1))
        atoms.set_atomic_numbers(atomicnumber*ones(len(atoms),int))
    else:
        atoms = None
    try:
        atoms = MakeParallelAtoms(atoms, cpuLayout)
        atoms.set_calculator(EMT())
        atoms.get_potential_energy()
    except Exception:
        print("*** SKIPPING TESTS - TOO FEW ATOMS/CPU ***")
        fasttable.write(("%-5s      "*2) % (("`-`",)*2))
        continue

    # Run the parallel timings (fast systems)
    MaxwellBoltzmannDistribution(atoms, 300 * units.kB)
    name = "Verlet%d-fast" % (atomspercpu,)
    atoms.set_calculator(EMT())
    dyn = VelocityVerlet(atoms, 5*units.fs)
    ParallelTiming(name, dyn.run, 4*steps, atoms.get_global_number_of_atoms(),
                   verlet, fasttable)
    name = "Langevin%d-fast" % (atomspercpu,)
    dyn = Langevin(atoms, 5*units.fs, 300*units.kB, 0.001)
    ParallelTiming(name, dyn.run, 4*steps, atoms.get_global_number_of_atoms(),
                   langevin, fasttable)
    #name = "NPT%d-fast" % (atomspercpu,)
    #dyn = HooverNPT(atoms, 5*units.fs, 300*units.kB, 0, 25*units.fs,
    #                (75*units.fs)**2 * 140*units.GPa)
    #ParallelTiming(name, dyn.run, 4*steps, atoms.get_global_number_of_atoms(),
    #               npt, fasttable)

if ismaster:
    bigtable.write("\n")
    fasttable.write("\n")
    
logger.close()
bigtable.close()
fasttable.close()



'''
Created on Mar 16, 2011

@author: schiotz
'''

from ase.md.md import MolecularDynamics
from ase.md.verlet import VelocityVerlet as _VelocityVerlet
from asap3.md.md import ParallelMolDynMixinNoData
import ase
import asap3
from ase.parallel import world
import numpy as np
import sys

class VelocityVerlet_Asap(MolecularDynamics, ParallelMolDynMixinNoData):
    def __init__(self, atoms, timestep, trajectory=None, logfile=None,
                 loginterval=1):
        MolecularDynamics.__init__(self, atoms, timestep, trajectory, logfile,
                                   loginterval)
        if not atoms.has('momenta'):
            atoms.set_momenta(np.zeros((len(atoms), 3), float))
        self.calculator = atoms.calc
        self.asap_md = asap3._asap.VelocityVerlet(atoms, self.calculator, timestep)
            
    def run(self, steps):
        assert(self.calculator is self.atoms.calc)
        self.before_run()
        # Extra stuff needs to be done for FixAtoms constraint.
        if self.atoms.constraints:
            if len(self.atoms.constraints) != 1:
                raise RuntimeError("AASAP Verlet can only do parallel dynamics with a single constraint.")
            constraint = self.atoms.constraints[0]
            if not  isinstance(constraint, asap3.constraints.FixAtoms):
                raise RuntimeError("ASAP Verlet only supports constrained dynamics with the ASAP FixAtoms constraint.")
            mask = constraint.index
            if not mask.dtype == bool:
                mask2 = np.zeros(len(self.atoms), bool)
                mask2[mask] = True
                mask = mask2
            assert mask.shape == (len(self.atoms),) and mask.dtype == bool
            mult = np.logical_not(mask).astype(float)
            self.atoms.arrays["FixAtoms_mult_double"] = mult
            del mask, mult
        self.asap_md.run(steps, self.observers, self)
        self.after_run()
        if self.atoms.constraints:
            del self.atoms.arrays["FixAtoms_mult_double"]

class VelocityVerlet_ASE(_VelocityVerlet, ParallelMolDynMixinNoData):
    def run(self, steps):
        self.before_run()
        super().run(steps)
        self.after_run()
        
def VelocityVerlet(atoms, timestep, trajectory=None, logfile=None, loginterval=1):
    if isinstance(atoms, ase.Atoms) and asap3.constraints.check_asap_constraints(atoms):
        if world.rank == 0:
            sys.stderr.write("Using Asap-optimized Verlet algorithm\n")
        return VelocityVerlet_Asap(atoms, timestep, trajectory, logfile, loginterval)
    else:
        if world.rank == 0:
            sys.stderr.write("Using ASE Verlet algorithm\n")
        return VelocityVerlet_ASE(atoms, timestep, trajectory, logfile, loginterval)




#! /usr/bin/env python

#PBS -N BigTiming
#PBS -l nodes=1:ppn=4:opteron285
#PBS -q small

from numpy import *
from asap3 import *
from asap3.md.verlet import VelocityVerlet
from asap3.md.langevin import Langevin
from ase.lattice.cubic import FaceCenteredCubic
from asap3.timing import report_timing
from asap3.memory import get_max_memory
import sys, pickle, time, subprocess, os, re
import numpy as np
from asap3.testtools import ReportTest

# cpu time:  time.clock().   Wall clock time: time.time()

#Verbose(1)

print_version(1)

usethread = (len(sys.argv) > 1 and
             (sys.argv[1] == "-t" or sys.argv[1] == "-T"))
if usethread:
    if sys.argv[1] == "-t":
        AsapThreads()
    else:
        AsapThreads(4)

host = subprocess.check_output("hostname").decode('utf-8').strip()

timesteps = 25
if usethread:
    dbfilename = "timing-thread-big.dat"
    logfilename = "timing-thread-big.log"
else:
    dbfilename = "timing-big.dat"
    logfilename = "timing-big.log"
selfcheckfilename = "timing-selfcheck-big.dat"
asapversion = get_version()
when = time.strftime("%a %d %b %Y %H:%M", time.localtime(time.time()))

randomstate = "randomstate.pickle"
if os.path.isfile(randomstate):
    np.random.set_state(pickle.load(open(randomstate, "rb")))
else:
    print("Saving random state for next call.")
    rndfile = open(randomstate, "wb")
    pickle.dump(np.random.get_state(), rndfile)
    rndfile.close()
    

#PrintVersion(1)
print("Running ASAP timing on "+host+".")
if re.match(r"^n\d\d\d.dcsc.fysik.dtu.dk$", host):
    print("    This is a d512 node on Niflheim.")
    fullhost = "niflheim-d512/%s" % (host.split(".")[0])
    host = "niflheim-d512"
elif re.match(r"^[stu]\d\d\d.dcsc.fysik.dtu.dk$", host):
    print("    This is an s50 node on Niflheim.")
    fullhost = "niflheim-s50/%s" % (host.split(".")[0])
    host = "niflheim-s50"
else:
    fullhost = host
print("Current time is "+when)
print("")

print("Preparing system")
initial = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                            size=(70, 70, 70),
                            symbol="Cu")
ReportTest("Number of atoms", len(initial), 1372000, 0)
r = initial.get_positions()
r.flat[:] += 0.14 * sin(arange(3*len(initial)))
initial.set_positions(r)

print("Running self-test.")
atoms = Atoms(initial)
atoms.set_calculator(EMT())
e = atoms.get_potential_energies()
f = atoms.get_forces()
if os.access(selfcheckfilename, os.F_OK):
    olde, oldf = pickle.load(open(selfcheckfilename, "rb"))
    de = max(fabs(e - olde))
    df = max(fabs(f.flat[:] - oldf.flat[:]))
    print("Maximal deviation:  Energy", de, "  Force", df)
    ReportTest("Max force error", df, 0.0, 1e-11)
    ReportTest("Max energy error", de, 0.0, 1e-11)
    del olde, oldf
else:
    print("WARNING: No self-check database found, creating it.")
    pickle.dump((e, f), open(selfcheckfilename, "wb"))
del e,f,atoms

ReportTest.Summary(exit=1)

print("Preparing to run Verlet dynamics.")
atoms = Atoms(initial)
atoms.set_calculator(EMT())
dynamics = VelocityVerlet(atoms, 5*units.fs)

print("Running Verlet dynamics.")
startcpu, startwall = time.clock(), time.time()
dynamics.run(timesteps)

vcpu, vwall = time.clock() - startcpu, time.time() - startwall
vfraction = vcpu/vwall
sys.stderr.write("\n")
print("Verlet dynamics done.")
atoms.get_calculator().print_memory()
del dynamics, atoms

print("Preparing to run Langevin dynamics.")
atoms = Atoms(initial)
atoms.set_calculator(EMT())
dynamics = Langevin(atoms, 5*units.fs, 400*units.kB, 0.001)

print("Running Langevin dynamics.")
startcpu, startwall = time.clock(), time.time()
dynamics.run(timesteps)

lcpu, lwall = time.clock() - startcpu, time.time() - startwall
lfraction = lcpu/lwall
sys.stderr.write("\n")
print("Langevin dynamics done.")
atoms.get_calculator().print_memory()
del dynamics, atoms

print("")
print("")
print("TIMING RESULTS:")
print(f"Verlet:   CPU time {vcpu:.2f}s  Wall clock time {vwall:.2f}s ({vfraction * 100:.0f}%)")
print(f"Langevin: CPU time {lcpu:.2f}s  Wall clock time {lwall:.2f}s ({lfraction * 100:.0f}%)")
print("")

key = (host, timesteps)
if os.access(dbfilename, os.F_OK):
    database = pickle.load(open(dbfilename, "rb"))
    founddb = 1
else:
    print("WARNING: No database with previous results ("+dbfilename+") - creating it.")
    database = {}
    founddb = 0
    
try:
    best = database[key]
except:
    isbetter = (1,1)
    best = {}
    print("No old data to compare with.")
else:
    print("Best timing so far on "+host+":")
    print("Verlet:   CPU time {:.2f}s  Wall clock time {:.2f}s ({:.0f}%)".format(best["vcpu"], best["vwall"], best["vfraction"] * 100))
    print("Obtained "+best["vwhen"])
    print(best["vversion"])
    print("")
    print("Langevin: CPU time {:.2f}s  Wall clock time {:.2f}s ({:.0f}%)".format(best["lcpu"], best["lwall"], best["lfraction"] * 100))
    print("Obtained "+best["lwhen"])
    print(best["lversion"])
    print("")
    vimp = (best["vcpu"] - vcpu) * 100.0 / best["vcpu"]
    limp = (best["lcpu"] - lcpu) * 100.0 / best["lcpu"]
    print("Improvement (CPU times):")
    print(f"  Verlet: {vimp:.1f}%   Langevin: {limp:.1f}%")
    print("")
    isbetter = (vimp > 0.0, limp > 0.0)

if (vfraction < 0.95 or lfraction < 0.95):
    print("These data are not quite reliable - got less than 95% of the CPU time.")
    print("Data will not be saved, even if they are best.")
    isbetter = (0,0)

if isbetter != (0,0):
    print("Saving improved result")
    value = best
    if isbetter[0]:
        best["vwhen"] = when
        best["vcpu"] = vcpu
        best["vwall"] = vwall
        best["vfraction"] = vfraction
        best["vversion"] = asapversion
    if isbetter[1]:
        best["lwhen"] = when
        best["lcpu"] = lcpu
        best["lwall"] = lwall
        best["lfraction"] = lfraction
        best["lversion"] = asapversion
    database[key] = best
    newfilename = dbfilename + ".new"
    bakfilename = dbfilename + ".bak"
    pickle.dump(database, open(newfilename, "wb"), 1)
    if founddb:
        os.rename(dbfilename, bakfilename)
    os.rename(newfilename, dbfilename)

now=time.strftime("%Y/%m/%d %H:%M")
version = asapversion.split()[2]
compiler = asapversion.split("'")[1]
logline = ( "%-20s %s tot=%-6.2f  %6.2f %6.2f %-2.0f  %6.2f %6.2f %-2.0f  mem=%.0f MB  ver %s %s"
            % (fullhost[:20], now, vcpu+lcpu, vcpu, vwall, vfraction*100,
               lcpu, lwall, lfraction*100, get_max_memory(),
               version, compiler))
print("\nLog line:")
print(logline)

logfile = open(logfilename, "a")
logfile.write(logline+"\n")
logfile.close()

report_timing()

print("Max memory:", get_max_memory())




"""Usage: python SpecialEMT.py [init]

Used to test if the EMT potential still produces the same energies,
forces and stresses as it used to.  It tests all elements with the
standard EMT parameters, including alloys, and also the special
parameters for Ruthenium and for CuMg and CuZr metallic glasses.

This has some overlap with test_potentials.py
"""


from asap3 import EMT, EMThcpParameters, EMTMetalGlassParameters
import pickle
from ase.lattice.cubic import FaceCenteredCubic
from ase.lattice.hexagonal import HexagonalClosedPacked
from ase.lattice.compounds import B2
from asap3.test.pytest_markers import ReportTest, serial
import numpy as np

import pytest

@pytest.fixture
def data(datadir):
    datafile = datadir / "SpecialEMT.pickle"
    return pickle.load(open(datafile, "rb"), encoding='latin1')

@pytest.fixture
def elements():
    return ["Al", "Ni", "Cu", "Pd", "Ag", "Pt", "Au"]

def check(a, key, data):
    print("Testing", key)
    expected_n = data[key+"_n"]
    expected_e = data[key+"_e"]
    expected_f = data[key+"_f"]
    e = a.get_potential_energies()
    f = a.get_forces().ravel()
    ReportTest(f"Number of atoms ({key})", len(a), expected_n, 0)
    ReportTest(f"Energies ({key})", max(np.fabs(e-expected_e)), 0.0, 1e-8)
    ReportTest(f"Forces ({key})", max(np.fabs(f-expected_f)), 0.0, 1e-8)

@serial
def test_ordinary_emt(elements, data):
    print("Checking single elements")
    for e in elements:
        atoms = FaceCenteredCubic(size=(4,4,4), symbol=e, pbc=False)
        atoms.calc = EMT()
        check(atoms, e, data)

@serial
@pytest.mark.core
def test_b2_alloys(elements, data):
    print("Checking B2 alloys")
    for i in range(1, len(elements)):
        e1 = elements[i]
        e2 = elements[i-1]
        atoms = B2(size=(4,4,4), symbol=(e1,e2), latticeconstant=3.5, pbc=False)
        atoms.calc = EMT()
        check(atoms, e1+e2, data)

@serial
@pytest.mark.core
def test_ruthenium(data):
    print("Checking Ruthenium")
    atoms = HexagonalClosedPacked(directions=[[1,0,-1,0],[0,1,-1,0],[0,0,0,1]],
                                size=(4,4,4), symbol="Ru", pbc=False)
    atoms.calc = EMT(EMThcpParameters())
    check(atoms, "Ru", data)

@serial
@pytest.mark.core
def test_metal_glass(data):
    print("Checking metallic glasses")
    for e1, e2 in (("Cu", "Mg"), ("Cu", "Zr")):
        atoms = B2(size=(4,4,4), symbol=(e1,e2), latticeconstant=3.5, pbc=False)
        atoms.calc = EMT(EMTMetalGlassParameters())
        check(atoms, e1+e2, data)

    



from asap3 import EMT, NeighborList, NeighborCellLocator, Atoms
from ase.lattice.cubic import FaceCenteredCubic
import ase.data
from asap3.test.pytest_markers import ReportTest, serial
import numpy as np

import pytest


def CheckLists(nblist, nblist2, name, natoms, count=None):
    "Run the tests on a half and a full neighbor list."
    print("Running tests on", name)
    print(f"Testing {name}: Lengths of lists.")
    if count:
        sum = 0
        for lst in nblist:
            sum += len(lst)
        ReportTest("Absolute length of list", sum, count * natoms, 0)

    sum1 = sum2 = 0
    for nb in enumerate(nblist):
        sum1 += len(nb)
    for nb in enumerate(nblist2):
        sum2 += len(nb)
        
    ReportTest("Equal total length of lists", sum2, sum1, 0)
    
    print(f"Testing {name}: List 1 atoms on list 2.")
    for i, nb in enumerate(nblist):
        for jj in nb:
            j = int(jj)
            assert (j in nblist2[i]) != (i in nblist2[j]), f'Atom {j} on list {i} (forward)'
    
    print(f"Testing {name}: List 2 atoms on list 1.")
    for i, nb in enumerate(nblist2):
        for jj in nb:
            j = int(jj)
            assert (j in nblist[i]) != (i in nblist[j]), f'Atom {j} on list {i} (reverse)'


@pytest.mark.slow
@serial
def test_cellneighborlocator():
    element = "Cu"
    latconst = ase.data.reference_states[ase.data.atomic_numbers[element]]['a']

    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,1],[0,0,1]], size=(9,7,5),
                            symbol=element, debug=0, pbc=(0,0,0))
    print("Symmetry:", atoms.get_pbc())
    atoms.calc = EMT()
    epot = atoms.get_potential_energy()

    nblist1 = NeighborList(latconst * 0.5 * (1/np.sqrt(2) + 1), atoms, 0.0)
    nblist2 = NeighborCellLocator(latconst * 0.5 * (1/np.sqrt(2) + 1), Atoms(atoms), 0.0)
    CheckLists(nblist1, nblist2, "nearest-neigbor lists (free)", len(atoms))

    atoms = Atoms(atoms, pbc=(1,1,1))
    print("Symmetry:", atoms.get_pbc())

    nblist1 = NeighborList(latconst * 0.5 * (1/np.sqrt(2) + 1), atoms, 0.0)
    nblist2 = NeighborCellLocator(latconst * 0.5 * (1/np.sqrt(2) + 1), Atoms(atoms), 0.0)
    CheckLists(nblist1, nblist2, "nearest-neigbor lists (periodic)", len(atoms), 6)

    nblist1 = NeighborCellLocator(4.98409, atoms, 0.0)
    nblist2 = NeighborList(4.98409, Atoms(atoms), 0.0)
    CheckLists(nblist1, nblist2, "long neigbor lists (periodic)", len(atoms), 21)

    atoms = Atoms(atoms, pbc=(0,0,0))
    print("Symmetry:", atoms.get_pbc())

    nblist = NeighborCellLocator(4.98409, atoms, 0.0)
    nblist2 = NeighborList(4.98409, Atoms(atoms), 0.0)
    CheckLists(nblist, nblist2, "long neigbor lists (free)", len(atoms))

    atoms1 = Atoms(atoms, pbc=(1,1,1))
    atoms2 = Atoms(atoms, pbc=(1,1,1))
    nblist = NeighborCellLocator(4.98409, atoms1)
    nblist2 = NeighborList(4.98409, atoms2)

    x0 = atoms1.get_positions()[2,0]
    for i in range(30):
        r = atoms1.get_positions()
        r[2,0] += 0.05
        atoms1.set_positions(r)
        print(r[2,0], end=' ')
        r = atoms2.get_positions()
        r[2,0] += 0.05
        print(r[2,0])
        atoms2.set_positions(r)
        nblist.check_and_update(atoms1)
        nblist2.check_and_update(atoms2)
        CheckLists(nblist, nblist2, "Translation step "+str(i), len(atoms))




import numpy as np
from nvt_module import main
from ase.md.andersen import Andersen
import ase

from asap3.test.pytest_markers import serial, parallel
from ase.parallel import world

import pytest

@pytest.mark.slow
@serial
@pytest.mark.skipif(ase.__version__ < '3.24.0', reason='ASE is too old')
def test_nvt_andersen():
    def anddynmaker(atoms, T0, dt, tau, rng, logint):
        # tau is the energy relaxation time.  The velocity relaxation time
        # should be the double.
        aprob = dt / tau
        return Andersen(atoms, dt, temperature_K=T0, andersen_prob=aprob, logfile='-', loginterval=logint, rng=rng)

    rng = np.random.default_rng(2718281828459045)   # Use a fixed seed for reproducability
    #rng = np.random.default_rng()
    main(anddynmaker, rng, sloppytime=True)



from asap3 import *
from asap3.md.verlet import VelocityVerlet
from ase.lattice.cubic import FaceCenteredCubic
from ase.io.trajectory import *
from ase.io import write
from numpy import *
import sys, os, time
from asap3.testtools import ReportTest

#DebugOutput("output.%d")
print_version(1)
delete = True
precision = 1e-8

def maketraj(atoms, t, nstep):
    e = [atoms.get_potential_energy()]
    dyn = VelocityVerlet(atoms, 5*units.fs)
    for i in range(nstep):
        dyn.run(10)
        e.append(atoms.get_potential_energy())
        if t is not None:
            t.write()
    return e

def maketraj_no_e(atoms, t, nstep):
    dyn = VelocityVerlet(atoms, 5*units.fs)
    for i in range(nstep):
        dyn.run(10)
        if t is not None:
            t.write()

def checktraj(t, e, recalc=True):
    i = 0
    for atoms in t:
        if recalc:
            atoms.calc = EMT()
        ReportTest("Checking frame %d" % (i,), atoms.get_potential_energy(),
                   e[i], precision)
        i += 1

initial = FaceCenteredCubic(size=(10,10,10), symbol="Cu", pbc=(1,0,0))


atoms = initial.copy()
atoms.calc = EMT()
print("Writing trajectory")
traj = Trajectory("traj1.traj", "w", atoms)
atoms.get_potential_energy()
traj.write()
energies = maketraj(atoms, traj, 10)
traj.close()

print("Reading trajectory (recalculating energies)")
traj = Trajectory("traj1.traj")
checktraj(traj, energies)

print("Reading trajectory (stored energies)")
traj = Trajectory("traj1.traj")
checktraj(traj, energies, recalc=False)

print("Writing a trajectory without precalculated energies.")
atoms = initial.copy()
atoms.calc = EMT()
traj = Trajectory("traj3.traj", "w", atoms, properties=['energy'])
atoms.get_potential_energy()
traj.write()
maketraj_no_e(atoms, traj, 10)
traj.close()
traj = Trajectory("traj3.traj")
checktraj(traj, energies, recalc=False)

if delete:
    print("Deleting trajectory")
    os.unlink("traj1.traj")
    os.unlink("traj3.traj")
    
ReportTest.Summary()



import random
import numpy as np
import math


def quaternion_to_rotation_matrix(q):
    [a, b, c, d] = q.tolist()

    return np.array([[a*a + b*b - c*c - d*d, 2*b*c - 2*a*d, 2*b*d + 2*a*c],\
            [2*b*c + 2*a*d, a*a - b*b + c*c - d*d, 2*c*d - 2*a*b],\
            [2*b*d - 2*a*c, 2*c*d + 2*a*b, a*a - b*b - c*c + d*d]])

def random_quaternion():

    x0 = random.uniform(0.0, 1.0)
    x1 = random.uniform(0.0, 2.0 * math.pi)
    x2 = random.uniform(0.0, 2.0 * math.pi)

    r1 = math.sqrt(1.0 - x0)
    r2 = math.sqrt(x0)

    s1 = math.sin(x1)
    c1 = math.cos(x1)
    s2 = math.sin(x2)
    c2 = math.cos(x2)
    return np.array([s1*r1, c1*r1, s2*r2, c2*r2])

def random_rotation_matrix():
    """Returns a random rotation matrix."""
    return quaternion_to_rotation_matrix(random_quaternion())

def rotation_matrix_to_quaternion(u):

    r11 = u[0][0]
    r12 = u[0][1]
    r13 = u[0][2]
    r21 = u[1][0]
    r22 = u[1][1]
    r23 = u[1][2]
    r31 = u[2][0]
    r32 = u[2][1]
    r33 = u[2][2]

    q = np.array([    1.0 + r11 + r22 + r33,
            1.0 + r11 - r22 - r33,
            1.0 - r11 + r22 - r33,
            1.0 - r11 - r22 + r33    ]).astype(np.double) / 4.0

    q = np.sqrt(np.maximum(0, q))
    i = np.argmax(q)

    if i == 0:
        q[1] *= np.sign(r32 - r23)
        q[2] *= np.sign(r13 - r31)
        q[3] *= np.sign(r21 - r12)

    elif i == 1:
        q[0] *= np.sign(r32 - r23)
        q[2] *= np.sign(r21 + r12)
        q[3] *= np.sign(r13 + r31)

    elif i == 2:
        q[0] *= np.sign(r13 - r31)
        q[1] *= np.sign(r21 + r12)
        q[3] *= np.sign(r32 + r23)

    elif i == 3:
        q[0] *= np.sign(r21 - r12)
        q[1] *= np.sign(r31 + r13)
        q[2] *= np.sign(r32 + r23)

    return q / np.linalg.norm(q)

HALF_SQRT_2 = 0.7071067811865474617150084668537601828575
generator_cubic = [    [1,    0,    0,    0    ],
            [0,    1,    0,    0    ],
            [0,    0,    1,    0    ],
            [0,    0,    0,    1    ],
            [0.5,    0.5,    0.5,    0.5    ],
            [0.5,    0.5,    -0.5,    0.5    ],
            [0.5,    -0.5,    0.5,    0.5    ],
            [0.5,    -0.5,    -0.5,    0.5    ],
            [-0.5,    0.5,    0.5,    0.5    ],
            [-0.5,    0.5,    -0.5,    0.5    ],
            [-0.5,    -0.5,    0.5,    0.5    ],
            [-0.5,    -0.5,    -0.5,    0.5    ],
            [HALF_SQRT_2,    HALF_SQRT_2,    0,    0    ],
            [HALF_SQRT_2,    0,    HALF_SQRT_2,    0    ],
            [HALF_SQRT_2,    0,    0,    HALF_SQRT_2    ],
            [-HALF_SQRT_2,    HALF_SQRT_2,    0,    0    ],
            [-HALF_SQRT_2,    0,    HALF_SQRT_2,    0    ],
            [-HALF_SQRT_2,    0,    0,    HALF_SQRT_2    ],
            [0,    HALF_SQRT_2,    HALF_SQRT_2,    0    ],
            [0,    HALF_SQRT_2,    0,    HALF_SQRT_2    ],
            [0,    0,    HALF_SQRT_2,    HALF_SQRT_2    ],
            [0,    -HALF_SQRT_2,    HALF_SQRT_2,    0    ],
            [0,    -HALF_SQRT_2,    0,    HALF_SQRT_2    ],
            [0,    0,    -HALF_SQRT_2,    HALF_SQRT_2    ]]

def quat_rot(r, a):
    b0 = r[0] * a[0] - r[1] * a[1] - r[2] * a[2] - r[3] * a[3]
    b1 = r[0] * a[1] + r[1] * a[0] + r[2] * a[3] - r[3] * a[2]
    b2 = r[0] * a[2] - r[1] * a[3] + r[2] * a[0] + r[3] * a[1]
    b3 = r[0] * a[3] + r[1] * a[2] - r[2] * a[1] + r[3] * a[0]
    return np.array([b0, b1, b2, b3]).astype(np.double)

def quat_map_cubic(r, i):
    return quat_rot(r, generator_cubic[i])

def rotate_quaternion_into_cubic_fundamental_zone(q):
    q = max([quat_map_cubic(q, i) for i in range(24)], key= lambda e: abs(e[0]))
    if q[0] < 0:
        q = -q
    return q

def quat_to_rodrigues(q):
    return np.array([q[1] / q[0], q[2] / q[0], q[3] / q[0]]).astype(np.double)

def rod_to_rgb(r):

    minh = math.radians(-62.8)
    maxh = math.radians(62.8)

    size = np.linalg.norm(r)
    r /= size
    theta = 2 * math.atan(size)
    return (r * theta - minh) / (maxh - minh)

def quat_to_rgb(q):

    rod = quat_to_rodrigues(q)
    return rod_to_rgb(rod)

def quat_inverse(q):
    return np.array([-q[0], q[1], q[2], q[3]])

def quat_quick_misorientation(q1, q2):

    t = np.dot(q1, q2)
    t = min(1, max(-1, t))
    return 2 * t * t - 1

def quat_misorientation(q1, q2):
    return math.acos(quat_quick_misorientation(q1, q2))




#!/usr/bin/env python
# Copyright (C) 2003-2020  CAMP
# Please see the accompanying LICENSE file for further information.

###### MISSING:
###
### * Verify recompilation on changes in both .cpp and .h files
### * Intel support
###
######


import os
import re
import runpy
import sys
from pathlib import Path
from subprocess import PIPE, run
from sysconfig import get_platform
from sysconfig import get_config_vars
import subprocess

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.develop import develop as _develop
from setuptools.command.install import install as _install

recordversion = runpy.run_path(Path(__file__).parent / 'recordversion.py')

# from config import build_interpreter, check_dependencies, write_configuration

# Get the current version number:
txt = Path('Python/asap3/version.py').read_text()
version = re.search("__version__ = '(.*)'", txt)[1]
ase_version_required = re.search("__ase_version_required__ = '(.*)'", txt)[1]

long_description = '''\
ASAP (Atomic SimulAtion Program or As Soon As Possible) is a
package for large-scale molecular dynamics within the Atomic
Simulation Environment (ASE).  It implements a number of 'classical'
potentials, most importantly the Effective Medium Theory, and also the
mechanisms for domain-decomposition of the atoms.'''


libraries = []
library_dirs = []
include_dirs = []
extra_link_args = []
extra_compile_args = ['-Wno-unknown-pragmas', '-Wno-sign-compare', '-Wno-unused-function', '-Wno-c++11-compat-deprecated-writable-strings', '-Wno-unknown-attributes']
runtime_library_dirs = []
extra_objects = []
define_macros = []
undef_macros = ['NDEBUG']

mpi_libraries = []
mpi_library_dirs = []
mpi_include_dirs = []
mpi_runtime_library_dirs = []
mpi_define_macros = []

compiler = None

folders = ['Basics', 'Potentials', 'Interface', 'Brenner', 'Tools',
           'PTM', 'PTM/qcprot', 'PTM/voronoi']
kim_folders = ['OpenKIMimport']
parallel_folders = ['Parallel', 'ParallelInterface']
exclude_files = ['Interface/AsapModule.cpp']
serial_only_files = ['Interface/AsapSerial.cpp']

# XXXX Move this to later, after identification of Intel compilers
if os.name != 'nt' and run(['which', 'mpicxx'], stdout=PIPE).returncode == 0:
    mpicompiler = 'mpicxx'
else:
    mpicompiler = None

mpilinker = mpicompiler

# Search and store current git hash if possible
try:
    from ase.utils import search_current_git_hash
    githash = search_current_git_hash('asap3')
    if githash is not None:
        define_macros += [('ASAP_GITHASH', githash)]
    else:
        print('.git directory not found. ASAP git hash not written.')
except ImportError:
    print('ASE not found. ASAP git hash not written.')

# User provided customizations:
asap_config = os.environ.get('ASAP_CONFIG')
if asap_config and not Path(asap_config).is_file():
    raise FileNotFoundError(asap_config)
for siteconfig in [asap_config,
                   'siteconfig.py']:
    if siteconfig is not None:
        path = Path(siteconfig).expanduser()
        if path.is_file():
            print('Reading configuration from', path)
            exec(path.read_text())
            break
else:  # no break
    # Make default configuration
    pass

# Try to use pkgconfig to locate OpenKIM
failed = subprocess.call("pkg-config --exists libkim-api",
                         shell=True)
if not failed:
    # pkg-config is installed and so is OpenKIM.
    kimincl = subprocess.check_output(
        "pkg-config --cflags-only-I libkim-api", shell=True).decode()
    include_dirs += [f.strip() for f in kimincl.split('-I') if f]
    include_dirs += kim_folders
    kimlibd = subprocess.check_output(
        "pkg-config --libs-only-L libkim-api", shell=True).decode()
    kimliblist = [f.strip() for f in kimlibd.split('-L') if f]
    library_dirs += kimliblist
    #runtime_library_dirs += kimliblist
    kimlibl = subprocess.check_output(
        "pkg-config --libs-only-l libkim-api", shell=True).decode()
    libraries += [f.strip() for f in kimlibl.split('-l') if f]
    kimother = subprocess.check_output(
        "pkg-config --libs-only-other libkim-api", shell=True).decode()
    if kimother.strip():
        extra_link_args += [kimother.strip()]
    define_macros += [('WITH_OPENKIM', '1')]
    folders.extend(kim_folders)

if mpicompiler:
    # Build MPI-interface into _gpaw.so:
    compiler = mpicompiler
    folders += parallel_folders
    exclude_files += serial_only_files

configvars = get_config_vars()
cmdclass = {}


# A hack so we don't try to import numpy before it is installed.  Adapted from
# https://stackoverflow.com/questions/19919905/how-to-bootstrap-numpy-installation-in-setup-py
class build_ext_using_numpy(build_ext):
    def finalize_options(self):
        super().finalize_options()
        import numpy
        self.include_dirs.append(numpy.get_include())
cmdclass['build_ext'] = build_ext_using_numpy

if compiler is not None:
    # A hack to change the used compiler and linker, inspired by
    # https://shwina.github.io/custom-compiler-linker-extensions/
    
    class custom_build_ext(build_ext_using_numpy):
        def build_extensions(self):
            # Override the compiler executables.
            for attr in ('compiler_so', 'compiler_cxx', 'compiler_so_cxx', 'linker_so'):
                if hasattr(self.compiler, attr):
                    temp = getattr(self.compiler, attr)
                    temp[0] = compiler
                    self.compiler.set_executable(attr, temp)
            super().build_extensions()
    cmdclass['build_ext'] = custom_build_ext

    
# Check if we are making a source distribution.  In that case, we
# should remove the VersionInfo_autogen folder and all files in it.
versiondir = 'VersionInfo_autogen'
is_making_distro = 'sdist' in sys.argv
if is_making_distro and os.path.isdir(versiondir):
    print('Clearing', versiondir)
    for f in os.listdir(versiondir):
        os.remove(os.path.join(versiondir, f))
    os.rmdir(versiondir)

# Create the version.cpp file
try:
    host = os.uname()[1]
except:
    host = 'unknown'
if not is_making_distro:
    versioninfo = f'{versiondir}/version_info.cpp'
    folders.append(versiondir)
    if not os.path.exists(versiondir):
        os.mkdir(versiondir)
    if len(os.listdir(versiondir)) > 1:
        print(f'WARNING: Unexpected files in {versiondir}.')
        for f in os.listdir(versiondir):
            cleanup = os.path.join(versiondir, f)
            print(f'    Cleaning up {cleanup}')
            os.remove(cleanup)
    try:
        myCC = os.environ['CXX']
    except KeyError:
        myCC = compiler
    try:
        myCFLAGS = os.environ['CXXFLAGS']
    except KeyError:
        myCFLAGS = ' '.join(extra_compile_args)
    if mpicompiler:
        compinfo = f'setuptools with {myCC} {myCFLAGS}'
        comptext = recordversion['contents'] % (version, 'parallel', host, compinfo, version)
    else:
        compinfo = f'setuptools with {myCC} {myCFLAGS}'
        comptext = recordversion['contents'] % (version, 'serial', host, compinfo, version)
    if os.path.exists(versioninfo):
        with open(versioninfo) as versioncpp:
            oldtext = versioncpp.read()
    else:
        oldtext = None
    if oldtext != comptext:
        print('Recording version info into VersionInfo_autogen')
        print('  CC =', myCC)
        print('  CFLAGS =', myCFLAGS)
        with open(versioninfo, 'w') as versioncpp:
            versioncpp.write(comptext)

# List source files
sources = []
for folder in folders:
    sources += Path(folder).glob('*.cpp')
for name in exclude_files:
    sources.remove(Path(name))
include_dirs += folders
# We cannot add numpy include dirs now, as numpy may not yet be installed.
### include_dirs.append(np.get_include())

# Make build process deterministic (for 'reproducible build')
sources = [str(source) for source in sources]
sources.sort()

## Check if this is necessary
## check_dependencies(sources)

# Convert Path objects to str:
library_dirs = [str(dir) for dir in library_dirs]
include_dirs = [str(dir) for dir in include_dirs]

extensions = [Extension('_asap',
                        sources,
                        libraries=libraries,
                        library_dirs=library_dirs,
                        include_dirs=include_dirs,
                        define_macros=define_macros,
                        undef_macros=undef_macros,
                        extra_link_args=extra_link_args,
                        extra_compile_args=extra_compile_args,
                        runtime_library_dirs=runtime_library_dirs,
                        extra_objects=extra_objects)]

# write_configuration(define_macros, include_dirs, libraries, library_dirs,
#                     extra_link_args, extra_compile_args,
#                     runtime_library_dirs, extra_objects, mpicompiler,
#                     mpi_libraries, mpi_library_dirs, mpi_include_dirs,
#                     mpi_runtime_library_dirs, mpi_define_macros)


# pyproject.toml does not have long_description
# Not sure how to handle platforms platforms=['unix']
setup(
      # extras_require={'docs': ['sphinx-rtd-theme',
      #                          'graphviz'],
      #                 'devel': ['flake8',
      #                           'mypy',
      #                           'pytest-xdist',
      #                           'interrogate']},
      ext_modules=extensions,
      cmdclass=cmdclass,
    )



"ListOfElements function returns which elements are present in an Atoms obj."

import numpy

def ListOfElements(atoms):
    """Return a list of the elements present in an Atoms object."""

    if getattr(atoms, "parallel", False):
        return atoms.get_list_of_elements()
    
    z = atoms.get_atomic_numbers()
    minz = min(z)
    maxz = max(z)
    loe = [minz]
    if minz != maxz:
        # More than one element
        for i in range (minz+1, maxz):
            if numpy.equal(z, i).any():
                loe.append(i)
        loe.append(maxz)
    return loe



import pytest
from ase.parallel import world
import asap3
import numpy as np

serial = pytest.mark.skipif(world.size > 1, reason='test is serial')
parallel = pytest.mark.skipif(world.size == 1, reason='test is parallel')
withmpi = pytest.mark.skipif(not asap3.parallelpossible, reason='Asap is not compiled with MPI support')
withOpenKIM = pytest.mark.skipif(not asap3.OpenKIMsupported, reason='Asap not compiled with OpenKIM')

# Default name of OpenKIM model is defined here, so it can be used outside fixtures.
_openkimmodel = 'EMT_Asap_Standard_JacobsenStoltzeNorskov_1996_AlAgAuCuNiPdPt__MO_115316750986_001'

def ReportTest(name, value, expected, atol):
    'A simple wrapper to avoid editing old tests.'
    assert np.isclose(value, expected, atol=atol), name



"""The potentials built into Asap."""

__docformat__ = "restructuredtext en"

from asap3.Internal.Builtins import _asap
from asap3.Internal.CheckArray import check_parameter_array, _ChkLJarray
import asap3

import numpy as np
import ase
import ase.calculators.emt
import ase.data
from ase.utils import basestring
from copy import copy
import sys

EMTParameters = _asap.EMTParameters
EMTDefaultParameters = _asap.EMTDefaultParameters
EMTRasmussenParameters = _asap.EMTRasmussenParameters
#EMTVariableParameters = _asap.EMTVariableParameters
#EMThcpParameters = _asap.EMThcpParameters

BrennerPotential = _asap.BrennerPotential

#IntVector = asap.IntVector
#DoubleVector = asap.DoubleVector


# class FullNeighborList(asap.NeighborList2):
#     "A full neighbor list object."
#     def __init__(self, atoms, *args):
#         asap.NeighborList2.__init__(self, atoms, *args)
#         self.EnableFullNeighborLists()
#         self.atoms = atoms  # Keep the atoms alive.
#         self.CheckAndUpdateNeighborList()

# class NeighborCellLocator(asap.NeighborCellLocator):
#     "A Neighbor list object."
#     def __init__(self, atoms, *args):
#         asap.NeighborCellLocator.__init__(self, atoms, *args)
#         self.atoms = atoms  # Keep the atoms alive.
#         self.CheckAndUpdateNeighborList()
    
# class NeighborList(asap.NeighborList2):
#     "A Neighbor list object."
#     def __init__(self, atoms, *args):
#         asap.NeighborList2.__init__(self, atoms, *args)
#         self.atoms = atoms  # Keep the atoms alive.
#         self.CheckAndUpdateNeighborList()

def get_cell_heights(atoms):
    "Get the heights of the unit cells (perpendicular to the surfaces)"
    heights = np.zeros(3)
    cell = atoms.get_cell()
    for i in range(3):
        direction = np.cross(cell[(i+1) % 3], cell[(i+2) % 3])
        heights[i] = abs(np.dot(cell[i], direction) / np.sqrt(np.dot(direction,direction)))
    return heights

def smallest_pbc_direction(atoms):
    smallest = 1.0e20  # Very big
    pbc = atoms.get_pbc()
    h = get_cell_heights(atoms)
    for i in range(3):
        if pbc[i] and h[i] < smallest:
            smallest = h[i]
    return smallest

class AsapPotential:
    """Mixin for all Asap potentials."""

    _stresscomp = np.array([[0, 5, 4], [5, 1, 3], [4, 3, 2]])

    def set_atoms_mixin(self, atoms):
        """Set the atoms.  This mixin fixes compatibility with ASE 3.18.0 and earlier."""
        if not hasattr(atoms, "get_global_number_of_atoms"):
            atoms.get_global_number_of_atoms = atoms.get_number_of_atoms

    def set_atoms(self, atoms, *args):
        """Set the atoms.  This default version fixes compatibility with ASE 3.18.0 and earlier."""
        self.set_atoms_mixin(atoms)
        # Now call the potential.
        super().set_atoms(atoms, *args)

    def get_stress(self, atoms):
        stress = self.get_virial(atoms)
        if not getattr(atoms, "_ase_handles_dynamic_stress", False):
            p = atoms.get_momenta()
            masses = atoms.get_masses()
            invmass = 1.0 / masses
            for alpha in range(3):
                for beta in range(alpha, 3):
                    stress[self._stresscomp[alpha,beta]] -= (p[:,alpha] * p[:,beta] * invmass).sum()
        stress /= atoms.get_volume()
        return stress

    def get_stresses(self, atoms):
        stresses = self.get_virials(atoms)
        if hasattr(self, 'get_atomic_volumes'):
            invvol = 1.0 / self.get_atomic_volumes()
        else:
            invvol = atoms.get_global_number_of_atoms() / atoms.get_volume()
        if not getattr(atoms, "_ase_handles_dynamic_stress", False):
            p = atoms.get_momenta()
            invmass = 1.0 / atoms.get_masses()
            for alpha in range(3):
                for beta in range(alpha, 3):
                    stresses[:,self._stresscomp[alpha,beta]] -= p[:,alpha] * p[:,beta] * invmass
        for i in range(6):
            stresses[:,i] *= invvol            
        return stresses

    def get_property(self, prop, atoms, allow_calculation=True):
        try:
            if (not allow_calculation and
                self.calculation_required(atoms, [prop])):
                return None
        except AttributeError:
            pass

        method = 'get_' + {'energy': 'potential_energy',
                           'magmom': 'magnetic_moment',
                           'magmoms': 'magnetic_moments',
                           'dipole': 'dipole_moment'}.get(prop, prop)
        try:
            result = getattr(self, method)(atoms)
        except AttributeError:
            raise asap3.PropertyNotImplementedError
        return result

    def get_name(self):
        return "asap3." + self._get_name()

    def get_neighborlist(self):
        """Return the neighbor locator"""
        return self._get_neighborlist()
    
    name = property(get_name, doc="Name of the potential")
    

class EMT(AsapPotential, _asap.EMT):
    """The Effective Medium Theory potential.

    Per default, the EMT potential uses standard parameters for the
    supported metals (Ni, Cu, Pd, Ag, Pt, Au).  If other parameters
    are desired, the relevant parameter provider can be provided as an
    optional argument.
    """
    def __init__(self, parameters=None, minimum_image=None, verbose=None):
        if parameters is None:
            parameters = EMTDefaultParameters()
        if verbose is None:
            verbose = _asap.verbose
        _asap.EMT.__init__(self, parameters, verbose)
        self.verbose = verbose
        self.absolute_max_cutoff = parameters.get_max_cutoff_beforeinit()
        self._set_atoms_called = False
        self.use_minimum_image = minimum_image
        
    def set_atoms(self, atoms, *args):
        self.set_atoms_mixin(atoms)
        if not self._set_atoms_called:
            if self.use_minimum_image is None:
                self.use_minimum_image = smallest_pbc_direction(atoms) >= 2.1 * self.absolute_max_cutoff
            if not self.use_minimum_image:
                #sys.stderr.write("Asap.EMT: Disabling minimum image convention.\n")
                self._use_imageatoms()
        self._set_atoms_called = True
        _asap.EMT.set_atoms(self, atoms, *args)

    def get_atomic_volumes(self):
        return self._get_atomic_volumes()

class MonteCarloEMT(AsapPotential, _asap.MonteCarloEMT):
    """The Effective Medium Theory potential optimized for Monte Carlo.

    Per default, the EMT potential uses standard parameters for the
    supported metals (Ni, Cu, Pd, Ag, Pt, Au).  If other parameters
    are desired, the relevant parameter provider can be provided as an
    optional argument.

    This version is optimized for Monte Carlo simulations, at the
    price of somewhat slower ordinary energy and force calculations
    and a larger memory footprint.  In addition, this version cannot
    be parallelized.
    """
    def __init__(self, parameters=None, verbose=None):
        if parameters is None:
            parameters = EMTDefaultParameters()
        if verbose is None:
            verbose = _asap.verbose
        self.verbose = verbose
        _asap.MonteCarloEMT.__init__(self, parameters, verbose)

class EMT2013(AsapPotential, _asap.EMT2013):
    """The Effective Medium Theory version 2011 potential.
    
    Parameters must be provided as a dictionary, the keys are
    elements (atomic numbers or strings), the values are dictionaries
    with parameters.  The names of the parameters are 'eta2', 'lambda',
    'kappa', 'E0', 'V0', 'S0' and 'n0'.
    
    If no_new_elements is set to True, the user promises that no
    new elements are introduced after the first time the calculator
    is attached to an Atoms object.  Only elements present at that
    instance will be initialized.
    
    If no_new_elements is False (the default), all elements present
    in the parameters dictionary will be initialized.  This may cause
    a performance penalty as neighbor lists will be large enough to
    accommodate the largest atoms supported.
    """
    def __init__(self, parameters, no_new_elements=False, minimum_image=None, verbose=None):
        try:
            par_ver = parameters['Calculator']
        except KeyError:
            raise ValueError("Dictionary with parameters should contain a" +
                             " Calculator item - be careful not to use an " +
                             "obsolete parameter definition !!")
        if par_ver != "EMT2013_V1":
            raise ValueError("The parameters appear to be for a calculator " +
                             "of type " + par_ver + 
                             " (expected EMT2013_V1).")
        params = {}
        max_s0 = 0.0
        for k in parameters.keys():
            if k == "Calculator":
                continue  # Skip the version specification.
            if isinstance(k, basestring):
                z = ase.data.atomic_numbers[k]
            else:
                z = k
            params[z] = copy(parameters[k])
            params[z]['mass'] = ase.data.atomic_masses[z]
            s0 = params[z]['S0']
            if s0 > max_s0:
                max_s0 = s0
        if verbose is None:
            verbose = _asap.verbose
        self.verbose = verbose
        _asap.EMT2013.__init__(self, params, no_new_elements, verbose)
        beta = 1.809399790563555  # ((16*pi/3)^(1./3.))/sqrt(2)
        self.maxcut = 0.5 * (np.sqrt(1.5) + np.sqrt(2.)) * np.sqrt(2.) * beta * max_s0 * 1.05
        self._set_atoms_called = False
        self.use_minimum_image = minimum_image
        
    def set_atoms(self, atoms, *args):
        self.set_atoms_mixin(atoms)
        if not self._set_atoms_called:
            if self.use_minimum_image is None:
                self.use_minimum_image = smallest_pbc_direction(atoms) >= 2.5 * self.maxcut
            if not self.use_minimum_image:
                #sys.stderr.write("Asap.EMT2013: Disabling minimum image convention.\n")
                self._use_imageatoms()
        self._set_atoms_called = True
        _asap.EMT2013.set_atoms(self, atoms, *args)

    def get_atomic_volumes(self):
        return self._get_atomic_volumes()


def EMT2011(parameters):
    """EMT2011 has been renamed EMT2013, please use the new version."""
    sys.stderr.write("\nASAP Warning: EMT2011 is deprecated, use EMT2013 instead.\n")
    return EMT2013(parameters)

class RGL(AsapPotential, _asap.RGL):
    """
    Calculator based on the RGL/Gupta semi empirical tight-binding potential.

    Parameters
    ----------
    elements: List of the elements that will be supported.

    The parameters can be given here as a dictionary with the form:
        {'Pt': [p, q, a, xi, r0], ('Pt','Y'): [...], ('Y','Y'): [...]}.

    p, q, a, xi, r0 (optional): The potential parameters listed in 2D arrays
    with each dimension equal to the number of elements. Only the upper
    triangular part is used. If there is only one element, then the
    parameters may be numbers.

    cutoff (optional): The distance at which the cutoff function should
    start given in nearest neighor distances. Default is 1.73.

    delta (optional): The length of the cutoff function given in nearest
    neighbor distances. Defauls is 0.15.

    The cutoff defaults are set to include the 3rd and not the 4th nearest
    neighors. The expected nearest neighbor distance is calculated based
    on the parameters.
    """

    def __init__(self, elements, p=None, q=None, a=None, xi=None, r0=None,
                 cutoff=1.73, delta=0.15, debug=False, verbose=None):
        # Check if parameters is given as a dictionary
        if isinstance(elements, dict):
            parameters = elements.copy()

            # Get elements
            elements = []
            elementmap = {}
            for key in parameters.keys():
                if isinstance(key, basestring):
                    symbols = [key]
                elif isinstance(key, tuple) and len(key) == 2:
                    symbols = list(key)
                else:
                    raise KeyError('Key must either be a string or a tuple with ' +
                                   'two elements.')
                for s in symbols:
                    if not s in elements:
                        elementmap[s] = len(elements)
                        elements.append(s)
            n = len(elements)

            # Get parameters
            p = np.zeros((n, n)) 
            q = np.zeros((n, n)) 
            a = np.zeros((n, n)) 
            xi = np.zeros((n, n)) 
            r0 = np.zeros((n, n)) 

            visit = np.zeros(n)
            for key in parameters.keys():
                if isinstance(key, basestring):
                    i = j = elementmap[key]
                elif isinstance(key, tuple) and len(key) == 2:
                    i = elementmap[key[0]]
                    j = elementmap[key[1]]
                else:
                    raise KeyError('Key must either be a string or a tuple with ' +
                                   'two elements.')

                visit[i] += 1
                visit[j] += 1

                p[i,j] = parameters[key][0]
                q[i,j] = parameters[key][1]
                a[i,j] = parameters[key][2]
                xi[i,j] = parameters[key][3]
                r0[i,j] = parameters[key][4]

            if np.any(visit != n + 1):
                raise ValueError('For some elements there are either too few or ' +
                                 'too many parameters.')

        # Interpret elements
        if not isinstance(elements, (tuple, list, np.ndarray)):
            elements = [elements]
        n = len(elements)

        for i, symbol in enumerate(elements):
            if isinstance(symbol, basestring):
                elements[i] = ase.data.atomic_numbers[symbol]
        elements = np.array(elements)


        # Interpret parameters
        p = check_parameter_array(n, "p", p)
        q = check_parameter_array(n, "q", q)
        a = check_parameter_array(n, "a", a)
        xi = check_parameter_array(n, "xi", xi)
        r0 = check_parameter_array(n, "r0", r0)

        # Calculate cutoff
        a1 = (np.log(np.sqrt(12) * a * p / (xi * q)) / (p - q) + 1) * r0

        if not np.all((1.0 < a1) & (a1 < 10.0)):
            raise ValueError('Unreasonable parameters - the nearest neighbor ' +
                             f'distance span [{a1.min():.3f}, {a1.max():.3f}]')

        rcs = cutoff * a1.max()
        rcd = delta * a1.max()
        rce = rcs + rcd
        #print "RGL cutoff: %.4f-%.4f" % (rcs, rcd)

        # Calculate parameters for the cutoff function
        qf = -xi * np.exp(-q * (rcs / r0 - 1)) / rcd**3
        qd = q / r0 * rcd
        q5 = (12.0*qf - 6.0*qf*qd + qf*qd*qd) / (2.0 * rcd**2)
        q4 = (15.0*qf - 7.0*qf*qd + qf*qd*qd) / rcd
        q3 = (20.0*qf - 8.0*qf*qd + qf*qd*qd) / 2.0

        pf = -a * np.exp(-p * (rcs / r0 - 1)) / rcd**3
        pd = p / r0 * rcd
        p5 = (12.0*pf - 6.0*pf*pd + pf*pd*pd) / (2.0 * rcd**2)
        p4 = (15.0*pf - 7.0*pf*pd + pf*pd*pd) / rcd
        p3 = (20.0*pf - 8.0*pf*pd + pf*pd*pd) / 2.0

        if debug:
            print("RGL potential parameters")
            print("p:", p)
            print("q:", q)
            print("a:", a)
            print("xi:", xi)
            print("r0:", r0)
            print("rcs:", rcs)
            print("rce:", rce)
            print("q5:", q5)
            print("q4:", q4)
            print("q3:", q3)
            print("p5:", p5)
            print("p4:", p4)
            print("p3:", p3)

        if verbose is None:
            verbose = _asap.verbose
        self.verbose = verbose
        _asap.RGL.__init__(self, elements, p, q, a, xi * xi, r0, p3, p4, p5,
                           q3, q4, q5, rcs, rce, verbose)

Gupta = RGL

class LennardJones(AsapPotential, _asap.LennardJones):
    """Lennard-Jones potential.

    Parameters:
    
    elements:  Lists the elements that will be supported.
    
    epsilon and sigma: The LJ parameters.  2D arrays with each
    dimension equal to the number of elements.  Only the lower
    triangular part is used.  Or a 1D array obtained by flattening the
    corresponding 2D array (this possibility is deprecated, and incurs
    less runtime testing).  If there is only one element, then epsilon
    and sigma may be numbers.

    rCut:  The cutoff distance.  Default: XXX

    modified: Should the potential be shifted so no jump occurs at the
    cutoff.  Default: True.
    """
    def __init__(self, elements, epsilon, sigma, rCut=-1.0, modified=True, verbose=None):
        try:
            numelements = len(elements)
        except TypeError:
            numelements = 1
            elements = [elements]

        epsilon = _ChkLJarray(epsilon, numelements, "epsilon")
        sigma = _ChkLJarray(sigma, numelements, "sigma")
        masses = [ase.data.atomic_masses[z] for z in elements]

        if verbose is None:
            verbose = _asap.verbose
        self.verbose = verbose
        _asap.LennardJones.__init__(self, numelements, elements, epsilon, sigma,
                                    masses, rCut, modified, verbose)

    # Atomic volumes disabled in the C code.
    #def get_atomic_volumes(self):
    #    return self._get_atomic_volumes()


class RahmanStillingerLemberg(AsapPotential, _asap.RahmanStillingerLemberg):
    """Rahman Stillinger Lemberg potential (RSL2).
    """
    def __init__(self, D0, R0, y, a1, b1, c1, a2, b2, c2, a3, b3, c3, 
                 elements, rCut=1.0, verbose=None):
        try:
            numelements = len(elements)
        except TypeError:
            numelements = 1
            elements = [elements]

        D0 = _ChkLJarray(D0, numelements, "D0")
        R0 = _ChkLJarray(R0, numelements, "R0")
        y = _ChkLJarray(y, numelements, "y")
        a1 = _ChkLJarray(a1, numelements, "a1")
        b1 = _ChkLJarray(b1, numelements, "b1")
        c1 = _ChkLJarray(c1, numelements, "c1")
        a2 = _ChkLJarray(a2, numelements, "a2")
        b2 = _ChkLJarray(b2, numelements, "b2")
        c2 = _ChkLJarray(c2, numelements, "c2")
        a3 = _ChkLJarray(a3, numelements, "a3")
        b3 = _ChkLJarray(b3, numelements, "b3")
        c3 = _ChkLJarray(c3, numelements, "c3")
        masses = [ase.data.atomic_masses[z] for z in elements]
        if verbose is None:
            verbose = _asap.verbose
        self.verbose = verbose
        _asap.RahmanStillingerLemberg.__init__(self, numelements, 
                    D0, R0, y, a1, b1, c1, a2, b2, c2, a3, b3, c3, 
                    elements, masses, rCut, verbose)

# def Ewald(q, elements, rCut=1.0):
#     """Lennard-Jones potential.

#     Parameters:
    
#     elements:  Lists the elements that will be supported.
    
#     epsilon and sigma: The LJ parameters.  2D arrays with each
#     dimension equal to the number of elements.  Only the lower
#     triangular part is used.  Or a 1D array obtained by flattening the
#     corresponding 2D array (this possibility is deprecated, and incurs
#     less runtime testing).  If there is only one element, then epsilon
#     and sigma may be numbers.

#     rCut:  The cutoff distance.  Default: XXX

#     modified: Should the potential be shifted so no jump occurs at the
#     cutoff.  Default: True.
#     """
#     try:
#         numelements = len(elements)
#     except TypeError:
#         numelements = 1
#         elements = [elements]

#       q = _ChkLJarray(q, numelements, "q")
#     masses = [ase.data.atomic_masses[z] for z in elements]
    
#     return _asap.Ewald(numelements, q, elements, masses, rCut)

class MetalOxideInterface(AsapPotential, _asap.MetalOxideInterface):
    def __init__(self,
                 P, Q, A, xi, r0, RGL_cut, 
                 q, kappa, 
                 D, alpha, R0, 
                 a, b, f0, oxide_cut,
                 beta, gamma, interface_cut, verbose=None):
        if verbose is None:
            verbose = _asap.verbose
        self.verbose = verbose
        _asap.MetalOxideInterface.__init__(self, 
            P, Q, A, xi, r0, RGL_cut, 
            q, kappa, 
            D, alpha, R0, 
            a, b, f0, oxide_cut,
            beta, gamma, interface_cut, verbose)
        
        self.max_cutoff = max(RGL_cut, oxide_cut, interface_cut)
        self.atoms_intialized = False
        self._set_atoms_called = False

    def set_atoms(self, atoms, *args):
        self.set_atoms_mixin(atoms)
        if not self._set_atoms_called:
            if smallest_pbc_direction(atoms) < 2.05 * self.max_cutoff:
                print('Using image atoms')
                self._use_imageatoms()
            self._set_atoms_called = True
            
        if hasattr(atoms, 'ghosts'):
            atoms.ghosts['assignment'] = np.zeros((0,),dtype=np.int32)
        if hasattr(atoms, 'ghosts'):
            atoms.ghosts['monolayer'] = np.zeros((0,),dtype=np.int32)
        
        _asap.MetalOxideInterface.set_atoms(self, atoms, *args)

class MetalOxideInterface2(AsapPotential, _asap.MetalOxideInterface2):
    def __init__(self,
                P, Q, A, xi, r0, RGL_cut, 
                q, kappa, 
                D, alpha, R0, 
                a, b, f0, oxide_cut,
                E, rho0, l0,
                B, C,
                interface_cut, verbose=None):
        if verbose is None:
            verbose =_asap.verbose
        self.verbose = verbose
        _asap.MetalOxideInterface2.__init__(self, 
            P, Q, A, xi, r0, RGL_cut, 
            q, kappa, 
            D, alpha, R0, 
            a, b, f0, oxide_cut,
            E, rho0, l0,
            B, C,
            interface_cut, verbose)
        
        self.max_cutoff = max(RGL_cut, oxide_cut, interface_cut)
        self.atoms_intialized = False
        self._set_atoms_called = False

    def set_atoms(self, atoms, *args):
        self.set_atoms_mixin(atoms)
        if not self._set_atoms_called:
            if smallest_pbc_direction(atoms) < 2.05 * self.max_cutoff:
                print('Using image atoms')
                self._use_imageatoms()
            self._set_atoms_called = True
            
        if hasattr(atoms, 'ghosts'):
            atoms.ghosts['assignment'] = np.zeros((0,),dtype=np.int32)
        if hasattr(atoms, 'ghosts'):
            atoms.ghosts['monolayer'] = np.zeros((0,),dtype=np.int32)

        _asap.MetalOxideInterface2.set_atoms(self, atoms, *args)

class Morse(AsapPotential, _asap.Morse):
    """Lennard-Jones potential.

    Parameters:
    
    elements:  Lists the elements that will be supported.
    
    epsilon, alpha and rmin: The Morse potential parameters.
    2D arrays with each  dimension equal to the number of elements.
    Only the lower triangular part is used.  Or a 1D array obtained by
    flattening the corresponding 2D array (this possibility is deprecated,
    and incurs less runtime testing).  If there is only one element,
    then epsilon and sigma may be numbers.

    rCut:  The cutoff distance.  Default: XXX

    modified: Should the potential be shifted so no jump occurs at the
    cutoff.  Default: True.
    """
    def __init__(self, elements, epsilon, alpha, rmin, rCut=-1.0, modified=True, verbose=None):
        try:
            numelements = len(elements)
        except TypeError:
            numelements = 1
            elements = [elements]
        epsilon = _ChkLJarray(epsilon, numelements, "epsilon", "Morse")
        alpha = _ChkLJarray(alpha, numelements, "alpha", "Morse")
        rmin = _ChkLJarray(rmin, numelements, "rmin", "Morse")
        if verbose is None:
            verbose = _asap.verbose
        self.verbose = verbose

        _asap.Morse.__init__(self, elements, epsilon, alpha,
                             rmin, rCut, modified, verbose)


# Disable ase.EMT to prevent catastrophic loss of performance.
ase.calculators.emt.EMT.disabled = "Disabled by loading asap3."

# Register if an OpenKIMcalculator is available
OpenKIMsupported = hasattr(_asap, 'OpenKIMcalculator')



"""The ASAP version number is specified in this file.

This file sets the __version__ variable to ASAP's version number.
The variable is imported by the main Asap module.  Furthermore, this module
prints the version number when executed, this is used by the makefile.

The __ase_version_required__ variable specified the version of ASE this
version of Asap is designed to work with.  Some functionality may be broken
with earlier versions.  If newer versions break something it is a bug that
should be reported and fixed.
"""

__version__ = '3.13.10'
__ase_version_required__ = '3.23.0'


if __name__ == '__main__':
    print(__version__)
    



"""Test that the parallel neighbor list.

 Checks that the parallel neighbor list returns the same number of
 neighbors as the serial one, both when it originates from a 
 potential and when it is created from Python.
 
 The test really only makes sense in parallel, but can run in serial."""

from asap3 import EMT, MakeParallelAtoms, FullNeighborList
from asap3.analysis.localstructure import RestrictedCNA, CoordinationNumbers
from asap3.mpi import world
from ase.build import bulk
import numpy as np

from asap3.test.pytest_markers import ReportTest, parallel

import pytest

ismaster = world.rank == 0
isparallel = world.size != 1
nbl_cutoff = 4.5

def makesystem(layout):
    atoms = bulk('Cu')
    atoms = atoms.repeat((15, 15, 15))
    if layout:
        atoms = atoms.repeat(layout)
    atoms.set_pbc((True, False, True))
    #atoms.set_pbc((False, True, True))
    del atoms[87]   #Create point defect
    return atoms

@parallel
@pytest.mark.parametrize('usepot', (True, False))
def test_par_neighborlist(usepot, cpulayout):
    # Make serial nb list.
    seratoms = makesystem(cpulayout)
    sernblist = FullNeighborList(nbl_cutoff, seratoms)

    # Make parallel nb list
    if ismaster:
        atoms = makesystem(cpulayout)
    else:
        atoms = None
    if isparallel:
        atoms = MakeParallelAtoms(atoms, cpulayout)
    natoms = atoms.get_global_number_of_atoms()
    if usepot:
        atoms.calc = EMT()
        old_energy = atoms.get_potential_energy()
        print(atoms.calc.get_cutoff())
        print(len(atoms), atoms.get_global_number_of_atoms())
        #assert atoms.get_calculator().get_cutoff() >= nbl_cutoff
    
    print(f"Testing parallel neighbor locator (usepot = {usepot})")
    nblist = FullNeighborList(nbl_cutoff, atoms)
    print("Length of neighbor locator", len(nblist), len(atoms))
    print(nblist[-1])
    if isparallel:
        ids = atoms.get_ids()
    else:
        ids = range(len(atoms))
    maxlen = -1
    minlen = 1e100
    for i in range(len(atoms)):
        l = len(nblist[i])
        l_exp = len(sernblist[ids[i]])
        ReportTest(f"NBlist length (i={i}, id={ids[i]})", 
                   l, l_exp, 0)
        if l > maxlen:
            maxlen = l
        if l < minlen:
            minlen = l
        if l == l_exp:
            snbl = {n for n in sernblist[ids[i]]}
            for nb in nblist[i]:
                if nb < len(atoms):
                    assert ids[nb] in snbl

    print("Neighbor list length:", minlen, '-', maxlen)
    if usepot:
        atoms[1].position += np.array((1e-5,0,0))
        ReportTest("Energy", atoms.get_potential_energy(), old_energy, 1e-4)
 



#!/usr/bin/env python

"""OfficialTiming.py - Asap timing script.

Reports timing of various types of dynamics in usec/atom/timestep.
"""

from numpy import *
from asap3 import *
from asap3.md.verlet import VelocityVerlet
from asap3.md.langevin import Langevin
from asap3.md.npt import NPT
from asap3.md.velocitydistribution import MaxwellBoltzmannDistribution
from asap3.EMT2013Parameters import EMT2013_candidate_parameters
from ase.lattice.cubic import FaceCenteredCubic, BodyCenteredCubic
from ase.lattice.compounds import L1_2
import sys, time, commands, os, re
import numpy as np
from asap3.testtools import ReportTest

import asap3.Internal.Threads

from StringIO import StringIO
import sys
import commands
import time

nthreads = 1
if len(sys.argv) > 1 and sys.argv[1] == "-t":
    nthreads = AsapThreads()

if len(sys.argv) > 1 and sys.argv[1] == "-T":
    nthreads = AsapThreads(2, force=True)

host = commands.getoutput("hostname")
logfilename = "officialtiming.log"
when = time.strftime("%a %d %b %Y %H:%M", time.localtime(time.time()))
asapversion = get_version()

random.seed([42, 12345])

class Logger(StringIO):
    def __init__(self, filename):
        StringIO.__init__(self)
        self.logfile = filename
    def write(self, x):
        sys.stdout.write(x)
        StringIO.write(self, x)
    def close(self):
        outfile = open(self.logfile, "a")
        outfile.write(self.getvalue())
        outfile.close()
        StringIO.close(self)

logger = Logger(logfilename)

def Timing(name, func, timesteps, natoms):
    print("Preparing timing")
    func(2)
    print("Running timing:", name)
    startcpu, startwall = time.clock(), time.time()
    func(timesteps)
    cpu, wall = time.clock() - startcpu, time.time() - startwall
    fraction = 100.0 * cpu/wall
    print("Time:", cpu, "CPU seconds  ", wall, "Wall seconds") 
    cpu *= 1e6 / (timesteps * natoms)
    wall *= 1e6 / (timesteps * natoms)
    logger.write("%s: %.2f usec/atom/timestep CPU-time  %.2f Wall time (%.1f%%)\n"
                 % (name, cpu, wall, fraction))

def temperature(atoms):
    return 2.0/3.0 * atoms.get_kinetic_energy() / len(atoms) / units.kB

def MakeCu(T=300, size=(29,29,30), pot=None):
    print("Preparing", T, "K Copper system.")
    atoms = FaceCenteredCubic(directions=[[1,0,0],[0,1,0],[0,0,1]],
                              symbol='Cu', size=size)
    if pot is None:
        pot = EMT()
    atoms.set_calculator(pot)
    MaxwellBoltzmannDistribution(atoms, 2*T * units.kB)
    #dyn = VelocityVerlet(atoms, 5*units.fs)
    dyn = Langevin(atoms, 5*units.fs, T*units.kB, 0.05)
    dyn.run(50)
    print("Done.  Temperature =", temperature(atoms), \
          "K.  Number of atoms: ", len(atoms))
    return atoms

def MakeCu3Ni(T=300, pot=None):
    print("Preparing", T, "K NiCu3 system.")
    atoms = L1_2(directions=[[1,0,0],[0,1,0],[0,0,1]], symbol=('Ni', 'Cu'),
                 latticeconstant=3.61, size=(29,29,30))
    if pot is None:
        pot = EMT()
    atoms.set_calculator(pot)
    MaxwellBoltzmannDistribution(atoms, 2*T * units.kB)
    #dyn = VelocityVerlet(atoms, 5*units.fs)
    dyn = Langevin(atoms, 5*units.fs, T*units.kB, 0.05)
    dyn.run(50)
    print("Done.  Temperature =", temperature(atoms), \
          "K.  Number of atoms: ", len(atoms))
    return atoms

# Make sure that CPU speed is revved up.
dummy = L1_2(directions=[[1,0,0],[0,1,0],[0,0,1]], symbol=('Au', 'Cu'),
             latticeconstant=4.08, size=(10,10,10), debug=0)
dummy.set_calculator(EMT())
dyn = Langevin(dummy, 5*units.fs, 300*units.kB, 0.05)
dyn.run(10)
del dummy

logger.write(f"\n\nRunning on {host} {when}\n")
modelname = "Unknown CPU model name"
cpumhz = "Unknown CPU speed"
try:
    lines = open("/proc/cpuinfo").readlines()
    for line in lines:
        if line[:10] == "model name":
            modelname = line
            break
    for line in lines:
        if line[:7] == "cpu MHz":
            cpumhz = line
            break
except:
    print("Cannot get CPU info from /proc/cpuinfo")

logger.write(cpumhz)
logger.write(modelname)
logger.write(asapversion+"\n")
logger.write("Number of threads: " + str(nthreads)+"\n")

atoms300 = MakeCu()

atoms = Atoms(atoms300)
atoms.set_calculator(EMT())
dyn = VelocityVerlet(atoms, 5*units.fs)
Timing("Ver300", dyn.run, 50, len(atoms))

atoms = MakeCu(1000)
dyn = VelocityVerlet(atoms, 5*units.fs)
Timing("Ver1000", dyn.run, 50, len(atoms))

atoms = Atoms(atoms300)
atoms.set_calculator(EMT())
dyn = Langevin(atoms, 5*units.fs, 300*units.kB, 0.001)
Timing("Langevin", dyn.run, 50, len(atoms))

atoms = Atoms(atoms300)
atoms.set_calculator(EMT())
dyn = NPT(atoms, 5*units.fs, 300*units.kB, 0, 25*units.fs,
          (75*units.fs)**2 * 140*units.GPa)
Timing("NPT", dyn.run, 50, len(atoms))

atoms = Atoms(atoms300, pbc=(0,0,0))
atoms.set_calculator(EMT())
dyn = VelocityVerlet(atoms, 5*units.fs)
Timing("FreeBC", dyn.run, 50, len(atoms))

atoms = MakeCu3Ni()
atoms.set_calculator(EMT())
dyn = VelocityVerlet(atoms, 5*units.fs)
Timing("Cu3Ni", dyn.run, 50, len(atoms))

atoms = MakeCu(size=(6,6,7))
atoms.set_calculator(EMT())
dyn = VelocityVerlet(atoms, 5*units.fs)
Timing("Tiny", dyn.run, 5000, len(atoms))

# Make tests with EMT2013
atoms300 = MakeCu(pot=EMT2013(EMT2013_candidate_parameters))

atoms = Atoms(atoms300)
atoms.set_calculator(EMT2013(EMT2013_candidate_parameters))
dyn = VelocityVerlet(atoms, 5*units.fs)
Timing("EMT2013_Ver300", dyn.run, 50, len(atoms))

atoms = Atoms(atoms300, pbc=(0,0,0))
atoms.set_calculator(EMT2013(EMT2013_candidate_parameters))
dyn = VelocityVerlet(atoms, 5*units.fs)
Timing("EMT2013_FreeBC", dyn.run, 50, len(atoms))

atoms = MakeCu3Ni(pot=EMT2013(EMT2013_candidate_parameters))
atoms.set_calculator(EMT2013(EMT2013_candidate_parameters))
dyn = VelocityVerlet(atoms, 5*units.fs)
Timing("EMT2013_Cu3Ni", dyn.run, 50, len(atoms))

atoms = MakeCu(size=(6,6,7), pot=EMT2013(EMT2013_candidate_parameters))
atoms.set_calculator(EMT2013(EMT2013_candidate_parameters))
dyn = VelocityVerlet(atoms, 5*units.fs)
Timing("EMT2013_Tiny", dyn.run, 5000, len(atoms))


elements = [29]
epsilon  = [0.15]
sigma    = [2.34]
rcut = 0.5*(sqrt(3)+2)*1.09*sigma[0]

atoms = Atoms(atoms300)
atoms.set_calculator(LennardJones(elements, epsilon, sigma, rcut, True))
dyn = VelocityVerlet(atoms, 5*units.fs)
Timing("L-J", dyn.run, 50, len(atoms))

logger.close()



"""Test periodic boundary conditions."""

from asap3 import *
from asap3.analysis.localstructure import RestrictedCNA, CoordinationNumbers
from asap3.mpi import world
from ase.build import bulk
import numpy as np
from asap3.test.pytest_markers import ReportTest, parallel

import pytest


ismaster = world.rank == 0
isparallel = world.size != 1

pbc_list = [(True, True, False),
                (False, True, True),
                (True, False, True)]

def makesystem():
    atoms = bulk('Cu')
    atoms = atoms.repeat((25, 25, 25))
    return atoms

@parallel
@pytest.mark.core
def test_par_pbc(cpulayout):
    seratoms = makesystem()
    seratoms.set_pbc(pbc_list[-1])
    seratoms.calc = EMT()
    old_energy = seratoms.get_potential_energy()

    for pbc in pbc_list:
        # Make parallel nb list
        if ismaster:
            atoms = makesystem()
            atoms.set_pbc(pbc)
        else:
            atoms = None
        if isparallel:
            atoms = MakeParallelAtoms(atoms, cpulayout)
        natoms = atoms.get_global_number_of_atoms()
        atoms.calc = EMT()
        energy = atoms.get_potential_energy()
        ReportTest(f"PBC={str(pbc)}", energy, old_energy, 1e-6)

